"""Description of Pd class."""
__author__ = 'ikibalin'
__version__ = "2020_08_19"

import numpy
from typing import NoReturn

from cryspy.A_functions_base.function_2_crystallography_base import \
    calc_cos_ang

from cryspy.A_functions_base.matrix_operations import calc_m1_m2_m1t
from cryspy.A_functions_base.unit_cell import calc_matrix_t

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.B_parent_classes.cl_3_data import DataN
from cryspy.B_parent_classes.preocedures import take_items_by_class

from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import \
    DiffrnRadiation
from cryspy.C_item_loop_classes.cl_1_extinction import Extinction
from cryspy.C_item_loop_classes.cl_1_phase import PhaseL
from cryspy.C_item_loop_classes.cl_1_refln import ReflnL
from cryspy.C_item_loop_classes.cl_1_refln_susceptibility import \
    ReflnSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_refine_ls import RefineLs
from cryspy.C_item_loop_classes.cl_1_chi2 import Chi2
from cryspy.C_item_loop_classes.cl_1_range import Range
from cryspy.C_item_loop_classes.cl_1_setup import Setup
from cryspy.C_item_loop_classes.cl_1_texture import Texture, TextureL
from cryspy.C_item_loop_classes.cl_1_exclude import ExcludeL

from cryspy.C_item_loop_classes.cl_1_tof_background import TOFBackground
from cryspy.C_item_loop_classes.cl_1_tof_background_by_points import TOFBackgroundPointL
from cryspy.C_item_loop_classes.cl_1_tof_intensity_incident import \
    TOFIntensityIncident
from cryspy.C_item_loop_classes.cl_1_tof_meas import TOFMeasL
from cryspy.C_item_loop_classes.cl_1_tof_parameters import TOFParameters
from cryspy.C_item_loop_classes.cl_1_tof_peak import TOFPeakL
from cryspy.C_item_loop_classes.cl_1_tof_proc import TOFProcL
from cryspy.C_item_loop_classes.cl_1_tof_profile import TOFProfile

from cryspy.E_data_classes.cl_1_crystal import Crystal
# from cryspy.E_data_classes.cl_1_mag_crystal import MagCrystal


class TOF(DataN):
    """
    Time-of-flight powder diffraction (polarized or unpolarized neutrons, 1d).

    Data items in the TOF category record details about
    powder diffraction measurements by Time of FLight.

    Methods
    -------
        - calc_profile

    Attributes
    ----------
        - setup, tof_profile, phase, tof_background, tof_meas (mandatory)
        - diffrn_radiation, chi2, range, extinction,
          texture, exclude, tof_proc, tof_peak, refine_ls, refln_#phase_name
          refln_susceptibility_#phase_name (optional)
    """

    CLASSES_MANDATORY = (TOFParameters, TOFProfile, PhaseL, TOFMeasL)
    CLASSES_OPTIONAL = (TOFBackground, TOFBackgroundPointL, DiffrnRadiation, Chi2, Range,
                        Texture, TOFIntensityIncident, ExcludeL, TOFProcL,
                        TOFPeakL, RefineLs, ReflnL, ReflnSusceptibilityL, Setup)
    # CLASSES_INTERNAL = ()

    CLASSES = CLASSES_MANDATORY + CLASSES_OPTIONAL

    PREFIX = "tof"

    # default values for the parameters
    D_DEFAULT = {}

    def __init__(self, data_name=None, **kwargs) -> NoReturn:
        super(TOF, self).__init__()

        self.__dict__["items"] = []
        self.__dict__["data_name"] = data_name

        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def apply_constraints(self):
        """Apply constraints."""
        pass
        # if self.pd_meas is not None:
        #     self.pd_meas.apply_constraints()

    def plots(self):
        if self.is_attribute("tof_proc"):
            tof_proc = self.tof_proc
            fig_s, ax_s = tof_proc.plot_sum_diff()
            ax_s = fig_s.axes[0]
            ax_hkl = fig_s.axes[1]
            ax_s.set_title(self.data_name + " - "+ax_s.title.get_text())
            y_min_s, y_max_s = ax_hkl.get_ylim()
            y_dist_s = y_max_s-y_min_s
            y_step_s = 0.
            for item in self.items:
                if isinstance(item, TOFPeakL):
                    np_time = item.numpy_time
                    ax_hkl.plot(np_time, 0.*np_time+y_min_s -
                                y_step_s, "|", label=item.loop_name)
                    y_step_s += 0.05*y_dist_s
            res = []
            ax_s.legend(loc='upper right')
            res.append((fig_s, ax_s))
            return res
        elif self.is_attribute("tof_meas"):
            return self.tof_meas.plots()

    def get_dictionary(self):
        dict_tof = super(TOF, self).get_dictionary()

        tof_meas, range_, exclude = None, None, None

        l_obj = take_items_by_class(self, (Range, ))
        if len(l_obj) > 0:
            range_ = l_obj[0]

        l_obj = take_items_by_class(self, (ExcludeL, ))
        if len(l_obj) > 0:
            exclude = l_obj[0]

        l_obj = take_items_by_class(self, (TOFMeasL, ))
        if len(l_obj) > 0:
            tof_meas = l_obj[0]

        if tof_meas is not None:
            time = numpy.array(tof_meas.time, dtype=float)

            time_min = range_.time_min
            time_max = range_.time_max
            dict_tof["time_min"] = time_min
            dict_tof["time_max"] = time_max

            flag_in = numpy.logical_and(
                time >= time_min,
                time <= time_max)

            dict_tof["time"] = time[flag_in]
            if tof_meas.is_attribute("intensity_plus"):
                int_plus = numpy.array(
                    tof_meas.intensity_plus, dtype=float)[flag_in]
                s_int_plus = numpy.array(
                    tof_meas.intensity_plus_sigma, dtype=float)[flag_in]
                int_minus = numpy.array(
                    tof_meas.intensity_minus, dtype=float)[flag_in]
                s_int_minus = numpy.array(
                    tof_meas.intensity_minus_sigma, dtype=float)[flag_in]
                dict_tof["signal_exp_plus"] = numpy.stack(
                    [int_plus, s_int_plus], axis=0)
                dict_tof["signal_exp_minus"] = numpy.stack(
                    [int_minus, s_int_minus], axis=0)
            else:
                int_sum = numpy.array(tof_meas.intensity, dtype=float)[flag_in]
                s_int_sum = numpy.array(
                    tof_meas.intensity_sigma, dtype=float)[flag_in]
                dict_tof["signal_exp"] = numpy.stack(
                    [int_sum, s_int_sum], axis=0)

            time_in_range = time[flag_in]
            flag_exclude = numpy.zeros(time_in_range.shape, dtype=bool)
            if exclude is not None:
                for item_e in exclude.items:
                    flag_in_1 = numpy.logical_and(
                        time_in_range >= item_e.time_min,
                        time_in_range <= item_e.time_max)
                    flag_exclude = numpy.logical_or(flag_exclude, flag_in_1)
            dict_tof["excluded_points"] = flag_exclude

        return dict_tof

    def take_parameters_from_dictionary(self, ddict_diffrn, l_parameter_name: list = None, l_sigma: list = None):
        keys = ddict_diffrn.keys()
        if l_parameter_name is not None:
            parameter_label = [hh[0] for hh in l_parameter_name]
        else:
            parameter_label = []
        if "beam_polarization" in keys:
            self.diffrn_radiation.polarization = ddict_diffrn["beam_polarization"]

        if "flipper_efficiency" in keys:
            self.diffrn_radiation.efficiency = ddict_diffrn["flipper_efficiency"]

        if "phase_scale" in keys:
            hh = ddict_diffrn["phase_scale"]
            for i_item, item in enumerate(self.phase.items):
                item.scale = float(hh[i_item])

        if "phase_ig" in keys:
            hh = ddict_diffrn["phase_ig"]
            for i_item, item in enumerate(self.phase.items):
                item.igsize = float(hh[i_item])

        if "spectrum_incident_coefficients" in keys:
            tof_intensity_incident = self.tof_intensity_incident
            hh = ddict_diffrn["spectrum_incident_coefficients"]
            tof_intensity_incident.a0 = float(hh[0])
            tof_intensity_incident.a1 = float(hh[1])
            tof_intensity_incident.a2 = float(hh[2])
            tof_intensity_incident.a3 = float(hh[3])
            tof_intensity_incident.a4 = float(hh[4])
            tof_intensity_incident.a5 = float(hh[5])
            tof_intensity_incident.a6 = float(hh[6])
            tof_intensity_incident.a7 = float(hh[7])
            tof_intensity_incident.a8 = float(hh[8])

        if "zero" in keys:
            tof_parameters = self.tof_parameters
            setattr(tof_parameters, "zero", float(ddict_diffrn["zero"]))
            setattr(tof_parameters, "dtt1", float(ddict_diffrn["dtt1"]))
            if "dtt2" in keys:
                setattr(tof_parameters, "dtt2", float(ddict_diffrn["dtt2"]))
            if "zerot" in keys:
                setattr(tof_parameters, "zerot", float(ddict_diffrn["zerot"]))
            if "dtt1t" in keys:
                setattr(tof_parameters, "dtt1t", float(ddict_diffrn["dtt1t"]))
            if "dtt2t" in keys:
                setattr(tof_parameters, "dtt2t", float(ddict_diffrn["dtt2t"]))

        if "background_coefficients" in keys:
            tof_background = self.tof_background
            hh = ddict_diffrn["background_coefficients"]
            for i_hh in range(hh.size):
                setattr(tof_background, f"coeff{i_hh+1:}", float(hh[i_hh]))

        if "background_intensity" in keys:
            hh = ddict_diffrn["background_intensity"]
            for i_item, item in enumerate(self.tof_backgroundpoint.items):
                item.intensity = float(hh[i_item])

        if "profile_peak_shape" in keys:
            profile_peak_shape = ddict_diffrn["profile_peak_shape"]
            if profile_peak_shape == "pseudo-Voigt":
                profile_sigmas = ddict_diffrn["profile_sigmas"]
                profile_alphas = ddict_diffrn["profile_alphas"]
                profile_betas = ddict_diffrn["profile_betas"]
                profile_gammas = ddict_diffrn["profile_gammas"]
                tof_profile = self.tof_profile
                setattr(tof_profile, "alpha0", profile_alphas[0])
                setattr(tof_profile, "alpha1", profile_alphas[1])
                setattr(tof_profile, "beta0", profile_betas[0])
                setattr(tof_profile, "beta1", profile_betas[1])
                setattr(tof_profile, "sigma0", profile_sigmas[0])
                setattr(tof_profile, "sigma1", profile_sigmas[1])
                setattr(tof_profile, "sigma2", profile_sigmas[2])
                setattr(tof_profile, "gamma0", profile_gammas[0])
                setattr(tof_profile, "gamma1", profile_gammas[1])
                setattr(tof_profile, "gamma2", profile_gammas[2])
            elif profile_peak_shape == "Gauss":
                profile_sigmas = ddict_diffrn["profile_sigmas"]
                profile_alphas = ddict_diffrn["profile_alphas"]
                profile_betas = ddict_diffrn["profile_betas"]
                tof_profile = self.tof_profile
                setattr(tof_profile, "alpha0", profile_alphas[0])
                setattr(tof_profile, "alpha1", profile_alphas[1])
                setattr(tof_profile, "beta0", profile_betas[0])
                setattr(tof_profile, "beta1", profile_betas[1])
                setattr(tof_profile, "sigma0", profile_sigmas[0])
                setattr(tof_profile, "sigma1", profile_sigmas[1])
                setattr(tof_profile, "sigma2", profile_sigmas[2])
            elif profile_peak_shape == "type0m":
                profile_alphas = ddict_diffrn["profile_alphas"]
                profile_betas = ddict_diffrn["profile_betas"]
                profile_sigmas = ddict_diffrn["profile_sigmas"]
                profile_gammas = ddict_diffrn["profile_gammas"]
                profile_rs = ddict_diffrn["profile_rs"]
                tof_profile = self.tof_profile
                setattr(tof_profile, "alpha1", profile_alphas[0])
                setattr(tof_profile, "alpha2", profile_alphas[1])
                setattr(tof_profile, "beta00", profile_betas[0])
                setattr(tof_profile, "beta01", profile_betas[1])
                setattr(tof_profile, "beta10", profile_betas[2])
                setattr(tof_profile, "sigma0", profile_sigmas[0])
                setattr(tof_profile, "sigma1", profile_sigmas[1])
                setattr(tof_profile, "sigma2", profile_sigmas[2])
                setattr(tof_profile, "gamma0", profile_gammas[0])
                setattr(tof_profile, "gamma1", profile_gammas[1])
                setattr(tof_profile, "gamma2", profile_gammas[2])
                setattr(tof_profile, "r01", profile_rs[0])
                setattr(tof_profile, "r02", profile_rs[1])
                setattr(tof_profile, "r03", profile_rs[2])

        if len(parameter_label) > 0:
            for name, sigma in zip(l_parameter_name, l_sigma):
                if name[0] == "phase_scale":
                    self.phase.items[name[1][0]].scale_sigma = float(sigma)
                if name[0] == "phase_ig":
                    self.phase.items[name[1][0]].igsize_sigma = float(sigma)
                if name[0] == "beam_polarization":
                    self.diffrn_radiation.polarization_sigma = float(sigma)
                if name[0] == "flipper_efficiency":
                    self.diffrn_radiation.efficiency_sigma = float(sigma)
                if name[0] == "texture_g1":
                    self.texture.items[name[1][0]].g1_sigma = float(sigma)
                if name[0] == "texture_g2":
                    self.texture.items[name[1][0]].g2_sigma = float(sigma)
                if name[0] == "texture_axis":
                    ind_param, ind_a = name[1]
                    if ind_param == 0:
                        self.texture.items[ind_a].h_ax_sigma = float(sigma)
                    if ind_param == 1:
                        self.texture.items[ind_a].k_ax_sigma = float(sigma)
                    if ind_param == 2:
                        self.texture.items[ind_a].l_ax_sigma = float(sigma)
                if name[0] == "profile_sigmas":
                    if name[1] == 0:
                        self.tof_profile.sigma0_sigma = float(sigma)
                    elif name[1] == 1:
                        self.tof_profile.sigma1_sigma = float(sigma)
                    elif name[1] == 2:
                        self.tof_profile.sigma2_sigma = float(sigma)
                if name[0] == "profile_gammas":
                    if name[1] == 0:
                        self.tof_profile.gamma0_sigma = float(sigma)
                    elif name[1] == 1:
                        self.tof_profile.gamma1_sigma = float(sigma)
                    elif name[1] == 2:
                        self.tof_profile.gamma2_sigma = float(sigma)

        if (("signal_plus" in keys) and ("signal_minus" in keys)):
            tof_proc = TOFProcL()
            tof_proc.numpy_time = numpy.round(ddict_diffrn["time"], decimals=5)
            tof_proc.numpy_intensity_plus_net = numpy.round(
                ddict_diffrn["signal_plus"], decimals=5)
            tof_proc.numpy_intensity_minus_net = numpy.round(
                ddict_diffrn["signal_minus"], decimals=5)
            if "signal_exp_plus" in keys:
                tof_proc.numpy_intensity_plus = numpy.round(
                    ddict_diffrn["signal_exp_plus"][0], decimals=5)
                tof_proc.numpy_intensity_plus_sigma = numpy.round(
                    ddict_diffrn["signal_exp_plus"][1], decimals=5)
                tof_proc.numpy_intensity_minus = numpy.round(
                    ddict_diffrn["signal_exp_minus"][0], decimals=5)
                tof_proc.numpy_intensity_minus_sigma = numpy.round(
                    ddict_diffrn["signal_exp_minus"][1], decimals=5)
            else:
                tof_proc.numpy_intensity = numpy.round(
                    ddict_diffrn["signal_exp"][0], decimals=5)
                tof_proc.numpy_intensity_sigma = numpy.round(
                    ddict_diffrn["signal_exp"][1], decimals=5)
            tof_proc.numpy_intensity_bkg_calc = numpy.round(
                ddict_diffrn["signal_background"], decimals=5)
            tof_proc.numpy_excluded = ddict_diffrn["excluded_points"]
            tof_proc.numpy_to_items()
            self.tof_proc = tof_proc

        l_tof_peak = []
        l_refln = []
        for item_phase in self.phase.items:
            s_label = f"dict_in_out_{item_phase.label:}"
            if s_label in keys:
                dict_crystal = ddict_diffrn[s_label]
                dict_crystal_keys = dict_crystal.keys()
                if (("index_hkl" in dict_crystal_keys) and ("time_hkl" in dict_crystal_keys)):
                    index_hkl = dict_crystal["index_hkl"]
                    time_hkl = dict_crystal["time_hkl"]

                    tof_peak = TOFPeakL(loop_name=item_phase.label)
                    int_plus_max, int_minus_max = None, None
                    if "iint_plus_with_factors" in dict_crystal_keys:
                        # int_plus_max = dict_crystal["iint_plus_with_factors"]
                        int_plus_max = dict_crystal["iint_plus"]

                    if "iint_minus_with_factors" in dict_crystal_keys:
                        # int_minus_max = dict_crystal["iint_minus_with_factors"]
                        int_minus_max = dict_crystal["iint_minus"]

                    if "f_nucl":
                        refln = ReflnL(loop_name=item_phase.label)
                        refln.numpy_index_h = index_hkl[0]
                        refln.numpy_index_k = index_hkl[1]
                        refln.numpy_index_l = index_hkl[2]
                        refln.numpy_a_calc = dict_crystal["f_nucl"].real
                        refln.numpy_b_calc = dict_crystal["f_nucl"].imag
                        refln.numpy_to_items()
                        l_refln.append(refln)

                    if not((int_plus_max is None) and (int_minus_max is None)):
                        tof_peak.numpy_index_h = index_hkl[0]
                        tof_peak.numpy_index_k = index_hkl[1]
                        tof_peak.numpy_index_l = index_hkl[2]
                        tof_peak.numpy_index_mult = dict_crystal["multiplicity_hkl"]
                        tof_peak.numpy_time = numpy.round(time_hkl, decimals=3)
                        tof_peak.numpy_intensity_plus = int_plus_max
                        tof_peak.numpy_intensity_minus = int_minus_max
                        tof_peak.numpy_to_items()
                        l_tof_peak.append(tof_peak)

        if len(l_tof_peak) > 0:
            self.add_items(l_tof_peak)
        if len(l_refln) > 0:
            self.add_items(l_refln)

    # def estimate_background(self):
    #     tof_background = self.tof_background
    #     tof_proc = self.tof_proc
    #     tof_proc.estimate_background(tof_background)



import numpy
import scipy
import scipy.interpolate

from cryspy.A_functions_base.matrix_operations import calc_m1_m2_m1t, calc_m_v

from cryspy.A_functions_base.unit_cell import \
    calc_sthovl_by_unit_cell_parameters, calc_matrix_t, calc_eq_ccs_by_unit_cell_parameters

from cryspy.A_functions_base.structure_factor import \
    calc_f_nucl_by_dictionary, calc_sft_ccs_by_dictionary, \
    calc_index_hkl_multiplicity_in_range, calc_f_m_perp_ordered_by_dictionary

from cryspy.A_functions_base.integrated_intensity_powder_diffraction import \
    calc_powder_iint_2d_para, calc_powder_iint_2d_ordered, calc_powder_iint_2d_mix

from cryspy.A_functions_base.preferred_orientation import \
    calc_preferred_orientation_pd2d, calc_gamma_nu_for_textured_peaks

from cryspy.A_functions_base.powder_diffraction_const_wavelength import \
    calc_profile_pseudo_voight_2d, calc_lorentz_factor, calc_ttheta_phi_by_gamma_nu

from .rhochi_diffrn import get_flags

na = numpy.newaxis


def calc_background(gamma, nu, background_gamma, background_nu, background_intensity, flag_background_intensity: bool = False):
    # f = scipy.interpolate.interp2d( 
    #     background_gamma, background_nu, background_intensity.transpose(), kind="linear", fill_value=None) #FIXME: not sure about transpose
    # intensity = f(gamma, nu).transpose()
    # dder = {}
    # if flag_background_intensity:
    #     dder["background_intensity"] = None

    ga_p = background_gamma
    nu_p = background_nu
    signal_p = background_intensity
    ga = gamma

    if ga.min() < ga_p.min():
        ga_p = numpy.insert(ga_p, 0, ga.min(), axis=0)
        signal_p = numpy.insert(signal_p, 0, signal_p[0,:], axis=0)

    if ga.max() >= ga_p.max():
        ga_p = numpy.append(ga_p, ga.max()+0.001)
        signal_p = numpy.append(signal_p, signal_p[-1:,:], axis=0)

    if nu.min() < nu_p.min():
        nu_p = numpy.insert(nu_p, 0, nu.min(), axis=0)
        signal_p = numpy.insert(signal_p, 0, signal_p[:,0], axis=1)

    if nu.max() >= nu_p.max():
        nu_p = numpy.append(nu_p, nu.max()+0.001)
        signal_p = numpy.append(signal_p, signal_p[:,-1:], axis=1)

    ga_left = ga_p[:-1]
    ga_right = ga_p[1:]
    nu_left = nu_p[:-1]
    nu_right = nu_p[1:]

    flags_ga = numpy.logical_and(
        ga[:, na] >= ga_left[na, :], 
        ga[:, na] < ga_right[na, :])

    flags_nu = numpy.logical_and(
        nu[:, na] >= nu_left[na, :],
        nu[:, na] < nu_right[na, :])

    arg_ga_1 = numpy.argwhere(flags_ga)[:,1]
    arg_ga_2 = arg_ga_1 + 1
    arg_nu_1 = numpy.argwhere(flags_nu)[:,1]
    arg_nu_2 = arg_nu_1 + 1

    f_11 = signal_p[arg_ga_1[:, na],arg_nu_1[na, :]]
    f_12 = signal_p[arg_ga_1[:, na],arg_nu_2[na, :]]
    f_21 = signal_p[arg_ga_2[:, na],arg_nu_1[na, :]]
    f_22 = signal_p[arg_ga_2[:, na],arg_nu_2[na, :]]

    c_11 = (ga_p[arg_ga_2] - ga)[:, na] * (nu_p[arg_nu_2] - nu)[na, :]
    c_12 = (ga_p[arg_ga_2] - ga)[:, na] * (nu - nu_p[arg_nu_1])[na, :]
    c_21 = (ga - ga_p[arg_ga_1])[:, na] * (nu_p[arg_nu_2] - nu)[na, :]
    c_22 = (ga - ga_p[arg_ga_1])[:, na] * (nu - nu_p[arg_nu_1])[na, :]
    denom = (ga_p[arg_ga_2] - ga_p[arg_ga_1])[:, na] * (nu_p[arg_nu_2] - nu_p[arg_nu_1])[na, :]

    signal = (c_11 * f_11 + c_12 * f_12 + c_21 * f_21 + c_22 * f_22)/denom
    dder = {}
    if flag_background_intensity:
        dder_b = numpy.zeros(ga.shape + nu.shape + ga_p.shape + nu_p.shape, dtype=float)
        dder_b[:, :, arg_ga_1[:, na],arg_nu_1[na, :]] += (c_11/denom)[na, na, :, :]
        dder_b[:, :, arg_ga_1[:, na],arg_nu_2[na, :]] += (c_12/denom)[na, na, :, :]
        dder_b[:, :, arg_ga_2[:, na],arg_nu_1[na, :]] += (c_21/denom)[na, na, :, :]
        dder_b[:, :, arg_ga_2[:, na],arg_nu_2[na, :]] += (c_22/denom)[na, na, :, :]
        dder["background_intensity"] = dder_b
    return signal, dder 


def calc_chi_sq_for_pd2d_by_dictionary(
        dict_pd, dict_crystals, dict_in_out: dict = None, flag_use_precalculated_data: bool=False,
        flag_calc_analytical_derivatives: bool = False):
    """Calculate chi_sq for diffrn experiment.
    """
    if dict_in_out is None:
        flag_dict = False
        flag_use_precalculated_data = False
        dict_in_out_keys = []
    else:
        flag_dict = True
        dict_in_out_keys = dict_in_out.keys()
    dict_pd_keys = dict_pd.keys()

    phase_name = [hh["name"].lower() for hh in dict_crystals]

    excluded_points = dict_pd["excluded_points"]
    dict_in_out["excluded_points"] = excluded_points

    gamma = dict_pd["gamma"]
    nu = dict_pd["nu"]
    offset_gamma = dict_pd["offset_gamma"]
    offset_nu = dict_pd["offset_nu"]
    flags_offset_gamma = dict_pd["flags_offset_gamma"]
    flags_offset_nu = dict_pd["flags_offset_nu"]

    gamma_zs = gamma - offset_gamma
    nu_zs = nu - offset_nu

    dict_in_out["gamma"] = gamma_zs
    dict_in_out["nu"] = nu_zs

    wavelength = dict_pd["wavelength"]
    flags_wavelength = dict_pd["flags_wavelength"]

    magnetic_field = dict_pd["magnetic_field"]

    flag_polarized, flag_unpolarized = False, False
    if "signal_exp_plus" in dict_pd_keys:
        flag_polarized = True
    elif "signal_exp" in dict_pd_keys:
        flag_unpolarized = True

    if flag_polarized:
        beam_polarization = dict_pd["beam_polarization"]
        flipper_efficiency = dict_pd["flipper_efficiency"]

        flags_beam_polarization = dict_pd["flags_beam_polarization"]
        flags_flipper_efficiency = dict_pd["flags_flipper_efficiency"]
    elif flag_unpolarized:
        beam_polarization = numpy.array([0.,], dtype=float)
        flipper_efficiency = numpy.array([0.,], dtype=float)
        flags_beam_polarization = numpy.array([False,], dtype=bool)
        flags_flipper_efficiency = numpy.array([False,], dtype=bool)

    alpha_det = numpy.arccos(
        numpy.sin(nu_zs)[na, :]/numpy.sqrt(
            2-2*numpy.cos(gamma_zs)[:, na] * numpy.cos(nu_zs)[na, :]
        )
    ) 
    dict_in_out["alpha_detector"] = alpha_det

    flag_ttheta = flags_offset_gamma or flags_offset_nu
    ttheta_zs, phi_zs, dder_ttheta_zs, dder_phi_zs = calc_ttheta_phi_by_gamma_nu(
        gamma_zs[:, na], nu_zs[na, :], flag_gamma=flags_offset_gamma, flag_nu=flags_offset_nu)

    ttheta_min = calc_ttheta_phi_by_gamma_nu(gamma_zs.min(), 0., flag_gamma=False, flag_nu=False)[0]
    ttheta_max = calc_ttheta_phi_by_gamma_nu(gamma_zs.max(), nu_zs.max(), flag_gamma=False, flag_nu=False)[0]

    sthovl_min = numpy.sin(0.5*ttheta_min - numpy.pi/90.)/wavelength
    if sthovl_min < 0:
        sthovl_min = 0.00001
    sthovl_max = numpy.sin(0.5*ttheta_max + numpy.pi/90.)/wavelength
    
    background_gamma = dict_pd["background_gamma"]
    background_nu = dict_pd["background_nu"]
    background_intensity = dict_pd["background_intensity"]
    flags_background_intensity = dict_pd["flags_background_intensity"]

    flag_background_intensity = numpy.any(flags_background_intensity)
    if (flag_use_precalculated_data and ("signal_background" in dict_in_out) and not(flag_background_intensity)):
        signal_background = dict_in_out["signal_background"]
    else:
        signal_background, dder_s_bkgr = calc_background(gamma, nu, background_gamma, background_nu, background_intensity,
            flag_background_intensity= (flag_background_intensity and flag_calc_analytical_derivatives))
        dict_in_out["signal_background"] = signal_background

    pd_phase_name = dict_pd["phase_name"]
    pd_phase_scale = dict_pd["phase_scale"]
    pd_phase_resolution_parameters = dict_pd["phase_resolution_parameters"] # U_phase, V_phase, W_phase, X_phase, Y_phase
    pd_phase_ig = dict_pd["phase_ig"] # IG_phase
    flags_pd_phase_scale = dict_pd["flags_phase_scale"]
    flags_pd_phase_resolution_parameters = dict_pd["flags_phase_resolution_parameters"] # U_phase, V_phase, W_phase, X_phase, Y_phase
    flags_pd_phase_ig = dict_pd["flags_phase_ig"] # IG_phase

    resolution_parameters = dict_pd["resolution_parameters"] # U, V, W, X, Y
    asymmetry_parameters = dict_pd["asymmetry_parameters"] # p1, p2, p3, p4

    p_phi = dict_pd["resolution_phi_parameter"] 
    flag_p_phi = dict_pd["flags_resolution_phi_parameter"] 

    flags_resolution_parameters = dict_pd["flags_resolution_parameters"] 
    flags_asymmetry_parameters = dict_pd["flags_asymmetry_parameters"] 
    flag_asymmetry_parameters = numpy.any(flags_asymmetry_parameters)
    

    if "texture_name" in dict_pd_keys:
        flag_texture = True
        pd_texture_name = dict_pd["texture_name"]
        pd_texture_g1 = dict_pd["texture_g1"]
        pd_texture_g2 = dict_pd["texture_g2"]
        pd_texture_axis = dict_pd["texture_axis"]
        pd_flags_texture_g1 = dict_pd["flags_texture_g1"]
        pd_flags_texture_g2 = dict_pd["flags_texture_g2"]
        pd_flags_texture_axis = dict_pd["flags_texture_axis"]
    else:
        flag_texture = False

    lorentz_factor, dder_lf = calc_lorentz_factor(ttheta_zs, flag_ttheta=flag_ttheta)
    dict_in_out["lorentz_factor"] = lorentz_factor


    total_signal_plus = numpy.zeros_like(ttheta_zs)
    total_signal_minus = numpy.zeros_like(ttheta_zs)
    for p_name, p_scale, p_resolution, p_ig, flags_p_scale, flags_p_resolution, flags_p_ig in zip(pd_phase_name, 
            pd_phase_scale, pd_phase_resolution_parameters.transpose(), pd_phase_ig,
            flags_pd_phase_scale, flags_pd_phase_resolution_parameters.transpose(), flags_pd_phase_ig):
        p_name=p_name.lower()
        flag_phase_texture = False
        if flag_texture:
            ind_texture = numpy.argwhere(pd_texture_name==p_name)
            if ind_texture.shape[0] != 0:
                texture_g1 = pd_texture_g1[ind_texture[0]]
                texture_g2 = pd_texture_g2[ind_texture[0]]
                texture_axis = pd_texture_axis[:, ind_texture[0]]
                flag_phase_texture = True
                flags_texture_g1 = pd_flags_texture_g1[ind_texture[0]]
                flags_texture_g2 = pd_flags_texture_g2[ind_texture[0]]
                flags_texture_axis = pd_flags_texture_axis[:, ind_texture[0]]
        
        ind_phase = phase_name.index(p_name)
        dict_crystal = dict_crystals[ind_phase]
        dict_crystal_keys = dict_crystal.keys()
        dict_in_out_keys = dict_in_out.keys()
        if f"dict_in_out_{p_name:}" in dict_in_out_keys:
            dict_in_out_phase = dict_in_out[f"dict_in_out_{p_name:}"]
        else:
            dict_in_out_phase = {}
            dict_in_out[f"dict_in_out_{p_name:}"] = dict_in_out_phase

        dict_in_out_phase_keys = dict_in_out_phase.keys()

        if "reduced_symm_elems" in dict_crystal_keys:
            reduced_symm_elems  = dict_crystal["reduced_symm_elems"]
            translation_elems = dict_crystal["translation_elems"]
        elif "full_symm_elems" in dict_crystal_keys:
            full_symm_elems = dict_crystal["full_symm_elems"]
        elif "full_mcif_elems" in dict_crystal_keys:
            full_mcif_elems  = dict_crystal["full_mcif_elems"]

        unit_cell_parameters = dict_crystal["unit_cell_parameters"]
        flags_unit_cell_parameters = dict_crystal["flags_unit_cell_parameters"]
        flag_unit_cell_parameters = numpy.any(flags_unit_cell_parameters)

        if flag_unit_cell_parameters:
            sc_uc = dict_crystal["sc_uc"]
            v_uc = dict_crystal["v_uc"]
            unit_cell_parameters = numpy.dot(sc_uc, unit_cell_parameters) + v_uc            

        if (flag_use_precalculated_data and 
                ("index_hkl" in dict_in_out_phase_keys) and 
                ("multiplicity_hkl" in dict_in_out_phase_keys) and not(flag_unit_cell_parameters or flag_ttheta)):
            index_hkl = dict_in_out_phase["index_hkl"]
            multiplicity_hkl = dict_in_out_phase["multiplicity_hkl"]
        else:
            if flag_phase_texture:
                reduced_symm_elems_p1 = numpy.array([[0], [0], [0], [1], [1], [0], [0], [0], [1], [0], [0], [0], [1]], dtype=int)
                translation_elems_p1 = numpy.array([[0], [0], [0], [1]], dtype=int)
                index_hkl, multiplicity_hkl = calc_index_hkl_multiplicity_in_range(
                    sthovl_min, sthovl_max, unit_cell_parameters, reduced_symm_elems_p1, translation_elems_p1, False)
                # index_hkl = numpy.reshape(numpy.stack([index_hkl, -1*index_hkl], axis=2), (3, 2*index_hkl.shape[1]))
                # multiplicity_hkl = numpy.ones_like(index_hkl[0])
            else:
                if "reduced_symm_elems" in dict_crystal_keys:
                    centrosymmetry = dict_crystal["centrosymmetry"]
                    index_hkl, multiplicity_hkl = calc_index_hkl_multiplicity_in_range(
                        sthovl_min, sthovl_max, unit_cell_parameters, reduced_symm_elems, translation_elems, centrosymmetry)
                else:
                    translation_elems_p1 = numpy.array([[0], [0], [0], [1]], dtype=int)
                    index_hkl, multiplicity_hkl = calc_index_hkl_multiplicity_in_range(
                        sthovl_min, sthovl_max, unit_cell_parameters, full_mcif_elems[:13], translation_elems_p1, False)

            if (("index_hkl" in dict_in_out_phase_keys) and flag_use_precalculated_data):
                if index_hkl.shape != dict_in_out_phase["index_hkl"].shape:
                    flag_use_precalculated_data = False
                else: 
                    flag_use_precalculated_data = numpy.all(numpy.logical_and(dict_in_out_phase["index_hkl"], index_hkl))

            dict_in_out_phase["index_hkl"] = index_hkl
            dict_in_out_phase["multiplicity_hkl"] = multiplicity_hkl

        flag_sthovl_hkl = flag_unit_cell_parameters
        sthovl_hkl, dder_sthovl_hkl = calc_sthovl_by_unit_cell_parameters(index_hkl,
            unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)

        flag_ttheta_hkl = flag_sthovl_hkl or flags_wavelength
        ttheta_hkl = 2*numpy.arcsin(wavelength * sthovl_hkl)
        dict_in_out_phase["ttheta_hkl"] = ttheta_hkl

        f_nucl, dder_f_nucl = calc_f_nucl_by_dictionary(
            dict_crystal, dict_in_out_phase, flag_use_precalculated_data=flag_use_precalculated_data)
        flag_f_nucl = len(dder_f_nucl.keys()) > 0

        flag_para = False
        if "atom_para_index" in dict_crystal_keys:
            sft_ccs, dder_sft_ccs = calc_sft_ccs_by_dictionary(
                dict_crystal, dict_in_out_phase, flag_use_precalculated_data=flag_use_precalculated_data)
            flag_sft_ccs  = len(dder_sft_ccs.keys()) > 0
            flag_matrix_t = flag_unit_cell_parameters
            matrix_t, dder_matrix_t = calc_matrix_t(
                index_hkl, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)

            flag_tensor_sigma = flag_sft_ccs or flag_unit_cell_parameters
            tensor_sigma, dder_tensor_sigma = calc_m1_m2_m1t(matrix_t, sft_ccs, flag_m1=flag_sft_ccs, flag_m2=flag_unit_cell_parameters)
            flag_para = True

        flag_ordered = False
        if "atom_ordered_index" in dict_crystal_keys:
            f_m_perp_o_ccs, dder_f_m_perp_o_ccs = calc_f_m_perp_ordered_by_dictionary(
                dict_crystal, dict_in_out_phase, flag_use_precalculated_data=flag_use_precalculated_data)
            flag_f_m_perp_o =  len(dder_f_m_perp_o_ccs.keys()) > 0
            flag_ordered = True

            flag_matrix_t = flag_unit_cell_parameters
            matrix_t, dder_matrix_t = calc_matrix_t(
                index_hkl, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)
            f_m_perp_o, dder_f_m_perp_o = calc_m_v(matrix_t, f_m_perp_o_ccs, flag_m=flag_unit_cell_parameters, flag_v=flag_f_m_perp_o)
            

        if flag_para and not(flag_ordered):
            flag_iint_plus_minus = flag_f_nucl or flag_tensor_sigma or flags_beam_polarization or flags_flipper_efficiency
            if (("iint_plus" in dict_in_out_phase_keys) and ("iint_minu" in dict_in_out_phase_keys) and
                    flag_use_precalculated_data and not(flag_iint_plus_minus)):
                iint_plus, iint_minus = dict_in_out_phase["iint_plus"], dict_in_out_phase["iint_minus"]
            else:
                iint_plus, iint_minus, dder_plus, dder_minus = calc_powder_iint_2d_para(
                    f_nucl, tensor_sigma, beam_polarization, flipper_efficiency, magnetic_field,
                    alpha_det, dict_in_out_phase, 
                    flag_f_nucl=flag_f_nucl and flag_calc_analytical_derivatives,
                    flag_tensor_sigma=flag_tensor_sigma and flag_calc_analytical_derivatives,
                    flag_polarization=flags_beam_polarization and flag_calc_analytical_derivatives,
                    flag_flipper=flags_flipper_efficiency and flag_calc_analytical_derivatives)

        elif not(flag_para) and flag_ordered:
            flag_iint_plus_minus = flag_f_nucl or flag_f_m_perp_o or flags_beam_polarization or flags_flipper_efficiency
            if (("iint_plus" in dict_in_out_phase_keys) and ("iint_minus" in dict_in_out_phase_keys) and
                    flag_use_precalculated_data and not(flag_iint_plus_minus)):
                iint_plus, iint_minus = dict_in_out_phase["iint_plus"], dict_in_out_phase["iint_minus"]
            else:
                iint_plus, iint_minus, dder_plus, dder_minus = calc_powder_iint_2d_ordered(
                    f_nucl, f_m_perp_o, beam_polarization, flipper_efficiency,
                    alpha_det, dict_in_out_phase, 
                    flag_f_nucl=flag_f_nucl and flag_calc_analytical_derivatives,
                    flag_f_m_perp=flag_f_m_perp_o and flag_calc_analytical_derivatives,
                    flag_polarization=flags_beam_polarization and flag_calc_analytical_derivatives,
                    flag_flipper=flags_flipper_efficiency and flag_calc_analytical_derivatives)

        elif flag_para and flag_ordered:
            flag_iint_plus_minus = flag_f_nucl or flag_tensor_sigma or flag_f_m_perp_o or flags_beam_polarization or flags_flipper_efficiency
            if (("iint_plus" in dict_in_out_phase_keys) and ("iint_minu" in dict_in_out_phase_keys) and
                    flag_use_precalculated_data and not(flag_iint_plus_minus)):
                iint_plus, iint_minus = dict_in_out_phase["iint_plus"], dict_in_out_phase["iint_minus"]
            else:
                iint_plus, iint_minus, dder_plus, dder_minus = calc_powder_iint_2d_mix(
                    f_nucl, tensor_sigma, f_m_perp_o, beam_polarization, flipper_efficiency, magnetic_field,
                    alpha_det, dict_in_out_phase, 
                    flag_f_nucl=flag_f_nucl and flag_calc_analytical_derivatives,
                    flag_tensor_sigma=flag_tensor_sigma and flag_calc_analytical_derivatives,
                    flag_f_m_perp_ordered=flag_f_m_perp_o and flag_calc_analytical_derivatives,
                    flag_polarization=flags_beam_polarization and flag_calc_analytical_derivatives,
                    flag_flipper=flags_flipper_efficiency and flag_calc_analytical_derivatives)
        else:
            iint_plus = numpy.square(numpy.abs(f_nucl))
            iint_minus = numpy.square(numpy.abs(f_nucl))

        dict_in_out_phase["iint_plus"] = iint_plus
        dict_in_out_phase["iint_minus"] = iint_minus

        if flag_phase_texture:
            flag_texture_g1 = numpy.any(flags_texture_g1)
            flag_texture_g2 = numpy.any(flags_texture_g2)
            flag_texture_axis = numpy.any(flags_texture_axis)
            flag_hh = numpy.any([flag_texture_g1, flag_texture_g2, flag_texture_axis])
            if (flag_use_precalculated_data and 
                    ("preferred_orientation" in dict_in_out_phase_keys) and
                    not(flag_hh)):
                preferred_orientation = dict_in_out_phase["preferred_orientation"]
            else:
                preferred_orientation, dder_po = calc_preferred_orientation_pd2d(alpha_det,
                    index_hkl, texture_g1, texture_g2, texture_axis, unit_cell_parameters,
                    flag_texture_g1=flag_texture_g1 and flag_calc_analytical_derivatives,
                    flag_texture_g2=flag_texture_g2 and flag_calc_analytical_derivatives,
                    flag_texture_axis=flag_texture_axis and flag_calc_analytical_derivatives)
                dict_in_out_phase["preferred_orientation"] = preferred_orientation

                # it is not necessary calculations but it is better to now (gamma,nu)_max of peaks

                eq_axis = calc_eq_ccs_by_unit_cell_parameters(
                    texture_axis, unit_cell_parameters, flag_unit_cell_parameters=False)[0]
                if "eq_ccs" in dict_in_out_phase_keys:
                    eq_ccs = dict_in_out_phase["eq_ccs"]
                else:
                    eq_ccs = calc_eq_ccs_by_unit_cell_parameters(index_hkl, unit_cell_parameters=unit_cell_parameters, flag_unit_cell_parameters=False)[0]
                    dict_in_out_phase["eq_ccs"] = eq_ccs
                ttheta_hkl = dict_in_out_phase["ttheta_hkl"]
                gamma_hkl, nu_hkl = calc_gamma_nu_for_textured_peaks(
                    eq_axis, eq_ccs, ttheta_hkl, texture_g1)
                dict_in_out_phase["gamma_hkl"] = gamma_hkl
                dict_in_out_phase["nu_hkl"] = nu_hkl
        
        flag_rp = numpy.any(flags_p_resolution) or numpy.any(flags_resolution_parameters)

        hh = resolution_parameters + p_resolution
        u, v, w, x, y = hh[0], hh[1], hh[2], hh[3], hh[4]
        p_1, p_2, p_3, p_4 = asymmetry_parameters[0], asymmetry_parameters[1], asymmetry_parameters[2], asymmetry_parameters[3]

        flag_profile_pv = flag_rp or flag_asymmetry_parameters or flag_p_phi
        if flag_use_precalculated_data and ("profile_pv" in dict_in_out_keys) and not(flag_profile_pv):
            profile_pv = dict_in_out_phase["profile_pv"]
        else:
            profile_pv, dder_pv = calc_profile_pseudo_voight_2d(ttheta_zs, phi_zs, ttheta_hkl, u, v, w, p_ig, x, y,
                p_1, p_2, p_3, p_4, 
                p_phi,
                flag_ttheta=flag_ttheta, flag_ttheta_hkl=flag_ttheta_hkl, flag_u=flag_rp,
                flag_v=flag_rp and flag_calc_analytical_derivatives,
                flag_w=flag_rp and flag_calc_analytical_derivatives,
                flag_i_g=flags_p_ig and flag_calc_analytical_derivatives,
                flag_x=flag_rp and flag_calc_analytical_derivatives,
                flag_y=flag_rp and flag_calc_analytical_derivatives, 
                flag_p_1=flag_asymmetry_parameters and flag_calc_analytical_derivatives,
                flag_p_2=flag_asymmetry_parameters and flag_calc_analytical_derivatives,
                flag_p_3=flag_asymmetry_parameters and flag_calc_analytical_derivatives,
                flag_p_4=flag_asymmetry_parameters and flag_calc_analytical_derivatives,
                flag_p_phi=flag_p_phi and flag_calc_analytical_derivatives)
            dict_in_out_phase["profile_pv"] = profile_pv


        # flags_p_scale
        iint_m_plus = iint_plus * multiplicity_hkl
        iint_m_minus = iint_minus * multiplicity_hkl


        if flag_phase_texture:
            # 0.5 to have the same meaning for the scale factor as in FullProf
            signal_plus = 0.5 * p_scale * lorentz_factor * (profile_pv * iint_m_plus * preferred_orientation).sum(axis=2) # sum over hkl
            signal_minus = 0.5 * p_scale * lorentz_factor * (profile_pv * iint_m_minus * preferred_orientation).sum(axis=2) 

        else:
            signal_plus = 0.5 * p_scale * lorentz_factor * (profile_pv * iint_m_plus).sum(axis=2) 
            signal_minus = 0.5 * p_scale * lorentz_factor * (profile_pv * iint_m_minus).sum(axis=2) 

        dict_in_out_phase["signal_plus"] = signal_plus
        dict_in_out_phase["signal_minus"] = signal_minus
        total_signal_plus += signal_plus
        total_signal_minus += signal_minus

    dict_in_out["signal_plus"] = total_signal_plus
    dict_in_out["signal_minus"] = total_signal_minus

    if flag_polarized:
        signal_exp_plus = dict_pd["signal_exp_plus"]
        signal_exp_minus = dict_pd["signal_exp_minus"]

        dict_in_out["signal_exp_plus"] = signal_exp_plus
        dict_in_out["signal_exp_minus"] = signal_exp_minus
        flag_chi_sq_sum = dict_pd["flag_chi_sq_sum"]
        flag_chi_sq_difference = dict_pd["flag_chi_sq_difference"]

    elif flag_unpolarized:
        signal_exp = dict_pd["signal_exp"]
        dict_in_out["signal_exp"] = signal_exp
        flag_chi_sq_sum = True
        flag_chi_sq_difference = False

    chi_sq = 0.
    n_point = 0
    if flag_chi_sq_sum:
        in_points = numpy.logical_not(excluded_points)
        if flag_polarized:
            signal_exp_sum = signal_exp_plus[0, :] + signal_exp_minus[0, :]
            signal_sigma_sum = numpy.sqrt(numpy.square(signal_exp_plus[1, :]) + numpy.square(signal_exp_minus[1, :]))
        elif flag_unpolarized:
            signal_exp_sum = signal_exp[0, :]
            signal_sigma_sum = signal_exp[1, :]
        nan_points_sum = numpy.logical_not(numpy.isnan(signal_exp_sum))
        in_points = numpy.logical_and(in_points, nan_points_sum)

        total_signal_sum = (total_signal_plus + total_signal_minus)
        diff_signal_sum = signal_exp_sum - total_signal_sum - signal_background
        inv_sigma_sq_sum = numpy.square(1./signal_sigma_sum)
        chi_sq_sum = numpy.sum(numpy.square(diff_signal_sum[in_points])*inv_sigma_sq_sum[in_points])
        chi_sq += chi_sq_sum
        n_point += numpy.sum(in_points)

    if flag_chi_sq_difference:
        signal_exp_diff = signal_exp_plus[0, :] - signal_exp_minus[0, :]
        signal_sigma_diff = numpy.sqrt(numpy.square(signal_exp_plus[1, :]) + numpy.square(signal_exp_minus[1, :]))
        total_signal_diff = total_signal_plus - total_signal_minus
        nan_points_diff = numpy.logical_not(numpy.isnan(signal_exp_diff))

        chi_sq_diff = numpy.sum(numpy.square((signal_exp_diff - total_signal_diff)[nan_points_diff]/signal_sigma_diff[nan_points_diff]))
        chi_sq += chi_sq_diff
        n_point += numpy.sum(nan_points_diff)

    if numpy.isnan(chi_sq):
        chi_sq = 1e30

    flags_pd2d = get_flags(dict_pd)
    l_flags_crystal = [get_flags(dict_crystal) for dict_crystal in dict_crystals]

    dder_signal_pd2d_sum = {}
    dder_signal_pd2d_difference = {}

    if flag_calc_analytical_derivatives:
        if flag_background_intensity:
            dder_signal_pd2d_sum['background_intensity'] = dder_s_bkgr['background_intensity']

    dder_signal_pd2d_sum_keys = dder_signal_pd2d_sum.keys()
    dder_signal_pd2d_difference_keys = dder_signal_pd2d_difference.keys()

    l_dder_plus_p = []
    l_parameter_name = []
    for way, flags in flags_pd2d.items():
        if flag_calc_analytical_derivatives:
            name = way[0]
            if name in dder_signal_pd2d_sum_keys:
                dder_plus_p = dder_signal_pd2d_sum[name][:, :, flags]
            l_dder_plus_p.append(dder_plus_p)

        pd_type_name = dict_pd["type_name"]
        ind_1d = numpy.atleast_1d(numpy.argwhere(flags)) #.flatten()
        parameter_name = [(pd_type_name, ) + way + (tuple(ind_1d[ind,:]), ) for ind in range(ind_1d.shape[0])]
        l_parameter_name.extend(parameter_name)

    for flags_crystal, dict_crystal in zip(l_flags_crystal, dict_crystals):
        for way, flags in flags_crystal.items():
            crystal_type_name = dict_crystal["type_name"]
            ind_1d = numpy.atleast_1d(numpy.argwhere(flags)) #.flatten()
            parameter_name = [(crystal_type_name, ) + way + (tuple(ind_1d[ind,:]), ) for ind in range(ind_1d.shape[0])]
            l_parameter_name.extend(parameter_name)


    der_chi_sq = numpy.zeros((len(l_parameter_name), ), dtype=float) 
    dder_chi_sq = numpy.zeros((len(l_parameter_name), len(l_parameter_name)), dtype=float)
    if flag_calc_analytical_derivatives:
        if len(l_dder_plus_p) > 0:
            dder_plus_p = numpy.concatenate(l_dder_plus_p, axis=2)
            if flag_chi_sq_sum:
                der_chi_sq = (((-2.*diff_signal_sum*inv_sigma_sq_sum)[:, :, na] * dder_plus_p).sum(axis=1)).sum(axis=0)
                dder_chi_sq = ((2*dder_plus_p[:, :, :, na] * dder_plus_p[:, :, na, :]).sum(axis=1)).sum(axis=0)

    return chi_sq, n_point, der_chi_sq, dder_chi_sq, l_parameter_name
    


import numpy

from cryspy.A_functions_base.unit_cell import calc_m_m_by_unit_cell_parameters
from cryspy.A_functions_base.matrix_operations import calc_m1_m2_inv_m1, calc_m_v, calc_mm_as_m_q_inv_m, calc_mm_as_m1_m2_inv_m1, calc_det_m

def calc_sc_fract_sc_b(symm_elems, atom_fract_xyz):
    sc_fract = (symm_elems[4:13]).sum(axis=1)/symm_elems.shape[1]
    b_s = symm_elems[:3]/(symm_elems.shape[1]*numpy.expand_dims(symm_elems[3], axis=0))

    atom_fract_xyz = numpy.mod(atom_fract_xyz, 1)
    r_s_x = calc_m_v(symm_elems[4:13], atom_fract_xyz, flag_m=False, flag_v=False)[0]
    n_s = numpy.expand_dims(atom_fract_xyz, axis=1)-r_s_x-b_s
    sc_b = (b_s + n_s).sum(axis=1)/n_s.shape[1]

    # sc_b = ().sum(axis=1)
    # x_new = calc_m_v(symm_elems[4:13], atom_fract_xyz, flag_m=False, flag_v=False)[0]
    # n_s, x0 = numpy.divmod(x_new,1)
    # n_s = -n_s.sum(axis=1)/n_s.shape[1]
    # sc_b = sc_b + n_s
    return sc_fract, sc_b


def calc_sc_beta(symm_elems):
    """Calculate \beta_av. = 1/N_s R_s \beta R_s^T
    """
    r_11, r_12, r_13 = symm_elems[4], symm_elems[5], symm_elems[6]
    r_21, r_22, r_23 = symm_elems[7], symm_elems[8], symm_elems[9]
    r_31, r_32, r_33 = symm_elems[10], symm_elems[11], symm_elems[12]

    r_r_t = numpy.stack([
        numpy.stack([r_11**2, r_12**2, r_13**2, 2*r_11*r_12, 2*r_11*r_13, 2*r_12*r_13], axis=0), # 11
        numpy.stack([r_21**2, r_22**2, r_23**2, 2*r_21*r_22, 2*r_21*r_23, 2*r_22*r_23], axis=0), # 22
        numpy.stack([r_31**2, r_32**2, r_33**2, 2*r_31*r_32, 2*r_31*r_33, 2*r_32*r_33], axis=0), # 33
        numpy.stack([r_11*r_21, r_12*r_22, r_13*r_23, r_11*r_22 + r_12*r_21, r_11*r_23 + r_13*r_21, r_12*r_23 + r_13*r_22], axis=0), # 12
        numpy.stack([r_11*r_31, r_12*r_32, r_13*r_33, r_11*r_32 + r_12*r_31, r_11*r_33 + r_13*r_31, r_12*r_33 + r_13*r_32], axis=0), # 13
        numpy.stack([r_21*r_31, r_22*r_32, r_23*r_33, r_21*r_32 + r_22*r_31, r_21*r_33 + r_23*r_31, r_22*r_33 + r_23*r_32], axis=0)], axis=0 # 23
    )
    sc_beta = r_r_t.sum(axis=2)/r_11.shape[0]
    return sc_beta


def calc_sc_chi(symm_elems, unit_cell_parameters, flag_unit_cell_parameters: bool = False):
    """Calculate symmetry constraint matrix for susceptibility
    It is supposed that determinant of r matrix is 1
    """
    m_m, dder_m_m = calc_m_m_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=False)

    r_direct = symm_elems[4:13]

    r_ccs, dder_r_ccs = calc_m1_m2_inv_m1(m_m, r_direct, flag_m1=False, flag_m2=False)

    if symm_elems.shape[0] == 14:
        theta = numpy.expand_dims(symm_elems[13], axis=(0,1))
        # det_r_ccs = calc_det_m(r_ccs, flag_m=False)[0]
        # res = theta_s*det_r_ccs
    else:
        theta = numpy.expand_dims(numpy.ones_like(symm_elems[0]), axis=(0,1))

    res, dder_mm = calc_mm_as_m_q_inv_m(r_ccs, flag_m= flag_unit_cell_parameters)

    # if r_direct.shape[1] == 2:
    #     print("r_ccs: ", r_ccs)

    mm = (res*theta).sum(axis=2)/res.shape[2]

    sc_chi = numpy.stack([mm[0,:], mm[4,:], mm[8,:], mm[1,:], mm[2,:], mm[5,:]], axis=0)
    dder_sc_chi = {}
    return sc_chi, dder_sc_chi


def calc_sc_chi_full(symm_elems, unit_cell_parameters, flag_unit_cell_parameters:bool=False):
    """Calculate symmetry constraint matrix for susceptibility
    It is supposed that determinant of r matrix is 1
    """
    m_m, dder_m_m = calc_m_m_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=False)

    r_direct = symm_elems[4:13]
    if symm_elems.shape[0] == 14:
        theta = numpy.expand_dims(symm_elems[13], axis=(0,1))
    else:
        theta = numpy.expand_dims(numpy.ones_like(symm_elems[0]), axis=(0,1))
    r_ccs, dder_r_ccs = calc_m1_m2_inv_m1(m_m, r_direct, flag_m1=False, flag_m2=False)

    res, dder_mm = calc_mm_as_m1_m2_inv_m1(r_ccs, flag_m1= flag_unit_cell_parameters)

    sc_chi_full = (res*theta).sum(axis=2)/res.shape[2]
    dder_sc_chi_full = {}
    return sc_chi_full, dder_sc_chi_full




"""Pd2dBackground class."""
import numpy
import scipy
import scipy.interpolate
from typing import NoReturn, Union

from cryspy.B_parent_classes.cl_1_item import ItemN

def inversed_hessian_to_correlation(inv_hessian):
    """Calculate correlation matrix and sigmas for inversed hessian matrix."""
    np_sigma_sq = numpy.diagonal(inv_hessian)
    np_na = numpy.newaxis
    np_sigma = numpy.sqrt(numpy.abs(np_sigma_sq))
    np_sigma_sq_matrix = np_sigma[:, np_na] * np_sigma[np_na, :]
    np_correlation_matrix = inv_hessian / np_sigma_sq_matrix
    return np_correlation_matrix, np_sigma

class InversedHessian(ItemN):
    """Inversed Hessian matrix.

    In mathematics, the Hessian matrix or Hessian is a square matrix of
    second-order partial derivatives of a scalar-valued function, or scalar
    field. It describes the local curvature of a function of many variables.

    Attributes
    ----------
        - with_labels

    Internal Attributes
    -------------------
        - label, matrix, correlation_matrix

    Methods
    -------
        - get_variable_names
        - get_variable_by_name, set_variable_by_name
    """

    ATTR_MANDATORY_NAMES = ("with_labels", )
    ATTR_MANDATORY_TYPES = (str, )
    # ("matrix", "matrix", "matrix", "matrix")
    ATTR_MANDATORY_CIF = ("with_labels", )

    ATTR_OPTIONAL_NAMES = ()
    ATTR_OPTIONAL_TYPES = ()
    ATTR_OPTIONAL_CIF = ()

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ("label", "matrix", "correlation_matrix", "sigma")
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "inversed_hessian"

    def __init__(self, **kwargs) -> NoReturn:
        super(InversedHessian, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self):
        """Form object."""
        l_1 = (self.with_labels).strip().split("\n")
        l_label, l_matrix = [], []
        for line in l_1[0:]:
            l_1 = line.strip().split()
            l_label.append(str(l_1[0]))
            l_matrix.append(l_1[1:])

        np_label = numpy.array(l_label, dtype=str)
        arg_sort = numpy.argsort(np_label)
        np_label = np_label[arg_sort]
        np_matrix = numpy.array(l_matrix, dtype=float)
        np_matrix = np_matrix[arg_sort,:][:, arg_sort]
        np_correlation_matrix, np_sigma = inversed_hessian_to_correlation(np_matrix)
        
        self.set_labels(np_label)
        self.set_inversed_hessian(np_matrix)
        self.set_correlation_matrix(np_correlation_matrix)
        self.set_sigmas(np_sigma)
        # self.form_inversed_hessian()

    def set_labels(self, labels) -> NoReturn:
        """Form 2theta_phi_intensity from internal attributes."""
        np_label = numpy.array(labels, dtype=str)
        self.__dict__["label"] = np_label

    def set_inversed_hessian(self, matrix) -> NoReturn:
        """Form 2theta_phi_intensity from internal attributes."""
        self.__dict__["matrix"] = matrix

    def set_correlation_matrix(self, matrix) -> NoReturn:
        """Form 2theta_phi_intensity from internal attributes."""
        self.__dict__["correlation_matrix"] = matrix

    def set_sigmas(self, sigmas) -> NoReturn:
        """Form 2theta_phi_intensity from internal attributes."""
        self.__dict__["sigma"] = sigmas

    def form_inversed_hessian(self):
        f_label = self.is_attribute("label")
        f_hessian = self.is_attribute("matrix")
        f_correlation = self.is_attribute("correlation_matrix")
        f_sigma = self.is_attribute("sigma")
        if f_label:
            np_label = numpy.array(self.label, dtype=str)
            arg_sort = numpy.argsort(np_label)
            np_label = np_label[arg_sort]
            self.set_labels(np_label)

            if f_hessian:
                np_matrix = numpy.array(self.matrix, dtype=float)
                np_matrix = np_matrix[arg_sort, :][:, arg_sort]
                self.set_inversed_hessian(np_matrix)            
            if f_correlation:
                np_cmatrix = numpy.array(self.correlation_matrix, dtype=float)
                np_cmatrix = np_cmatrix[arg_sort, :][:, arg_sort]
                self.set_correlation_matrix(np_cmatrix)            
            if f_sigma:
                np_sigma = numpy.array(self.sigma, dtype=float)[arg_sort]
                self.set_sigmas(np_sigma)

        if f_label & f_hessian:
            ls_out = [f"{lab:7} "+" ".join([f"{val:.10f}" for val in hess])
                      for lab, hess in zip(self.label, self.matrix)]
            self.with_labels = "\n".join(ls_out)

        elif f_label & f_correlation & f_sigma:
            np_correlation = self.correlation_matrix
            np_sigma = self.sigma
            
            np_na = numpy.newaxis
            np_hessian = np_correlation*np_sigma[:, np_na]*np_sigma[np_na, :]
            self.set_inversed_hessian(np_hessian)
            ls_out = [f"{lab:7} "+" ".join([f"{val:.10f}" for val in hess])
                      for lab, hess in zip(self.label, self.matrix)]
            self.with_labels = "\n".join(ls_out)

    def report(self):
        if self.is_defined():
            self.form_object()
            np_correlation = self.correlation_matrix
            np_sigma = self.sigma
            np_label = self.label
            np_numbers = numpy.arange(1,np_label.size+1).astype(str)
            ls_out = ["# Sigmas and correlation matrix"]
            ls_out.append("|SIGMAS | Cor.Matrix:|"+"|".join(np_numbers) + "|")
            ls_out.append("|-----|-----|"+"|".join(["-----" for hh in
                                                    np_numbers]) + "|")
            
            for i1, lab, mat, sig in zip(range(len(np_numbers)), np_numbers,
                                         np_correlation, np_sigma):
                ls_out.append(f"|{sig:.5f}|{lab:}|"+"|".join([
                    f"{val:5.2f}" for i2, val in enumerate(mat)]) + "|")
            ls_out.append("\nParameters:")
            for numb, label in zip(np_numbers, np_label):
                ls_out.append(f" - {numb:}. {label:}")
            # ls_out.extend([f"{i_n+1:2} -- {lab:7}: {sig:.5f}" for i_n, lab, sig in zip(
            #     range(len(np_sigma)), np_label, np_sigma)])
            # ls_out.append("\nCorrelation matrix:")
            # ls_out.extend([f" {i_n+1:2}: "+" ".join([
            #     f"{val:5.2f}" for val in mat]) for i_n, mat in enumerate(np_correlation)])
            return "\n".join(ls_out)

    def report_html(self):
        ls_html = []
        if self.is_defined():
            self.form_object()
            np_correlation = self.correlation_matrix
            np_sigma = self.sigma
            np_label = self.label
            np_numbers = numpy.arange(1,np_label.size+1).astype(str)
            
            ls_html.append("<table>")
            ls_html.append("<tr><th>Number</th><th>Parameter</th><th>Sigma</th>")
            for numb, lab, sig in zip(np_numbers, np_label, np_sigma):
                ls_html.append(f"<tr><td>{numb:}.</td><td>{lab:}</td><td>{sig:.5f}</td></tr>")
            ls_html.append("</table>")
            ls_html.append("<br>")
            ls_html.append("<b>Correlation matrix:</b>")
            ls_html.append("<table>")
            ls_html.append("<th>Number</th>"+"".join(
                [f"<th>{lab:}</th>" for lab in np_numbers]) + "</tr>")
            for i1, lab, mat in zip(range(len(np_numbers)), np_numbers, np_correlation):
                ls_html.append(f"<tr><th>{lab:}</th>"+"".join([
                    f"<td bgcolor='#ff2200'>{val:5.2f}</td>" if ((abs(val)>0.7) & (i1 != i2 ))
                    else f"<td>{val:5.2f}</td>" for i2, val in enumerate(mat)]) + "</tr>")
            ls_html.append("</table>")

        return "".join(ls_html)
            



import numpy
from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class TOFPeak(ItemN):
    """Integrated intensities for time-of-flight experiments.

    This section contains peak information extracted from the
    measured or, if present, the processed diffractogram. Each
    peak in this table will have a unique label (see _pd_peak_id).
    The reflections and phases associated with each peak will be
    specified in other sections (see the _pd_refln_ and
    _pd_phase_ sections).

    Note that peak positions are customarily determined from the
    processed diffractogram and thus corrections for position
    and intensity will have been previously applied.

    Attributes:
        - index_h, index_k, index_l (mandatory)
        - index_mult, time, intensity_plus, intensity_minus, width_time
    """
    ATTR_MANDATORY_NAMES = ("index_h", "index_k", "index_l")
    ATTR_MANDATORY_TYPES = (int, int, int)
    ATTR_MANDATORY_CIF = ("index_h", "index_k", "index_l")

    ATTR_OPTIONAL_NAMES = ("index_mult", "time", "intensity_plus",
                           "intensity_minus", "width_time")
    ATTR_OPTIONAL_TYPES = (int, float, float, float, float)
    ATTR_OPTIONAL_CIF = ("index_mult", "time", "intensity_plus",
                         "intensity_minus", "width_time")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"time": "{:.2f}", "intensity_plus": "{:.2f}",
                 "intensity_minus": "{:.2f}", "width_time": "{:.5f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "tof_peak"

    def __init__(self, **kwargs) -> NoReturn:
        super(TOFPeak, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class TOFPeakL(LoopN):
    """Integrated intensities for time-of-flight experiments.

    This section contains peak information extracted from the
    measured or, if present, the processed diffractogram. Each
    peak in this table will have a unique label (see _pd_peak_id).
    The reflections and phases associated with each peak will be
    specified in other sections (see the _tof_refln_ and
    _phase_ sections).

    Note that peak positions are customarily determined from the
    processed diffractogram and thus corrections for position
    and intensity will have been previously applied.

    Attributes:
        - index_h, index_k, index_l (mandatory)
        - index_mult, time, intensity_plus, intensity_minus, width_time

    """
    ITEM_CLASS = TOFPeak
    ATTR_INDEX = None
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(TOFPeakL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name


# s_cont = """
#   loop_
#   _tof_peak_index_h
#   _tof_peak_index_k
#   _tof_peak_index_l
#   _tof_peak_index_mult
#   _tof_peak_time
#   _tof_peak_intensity_plus
#   _tof_peak_intensity_minus
#   _tof_peak_width_time
#   2  2  0  4 17.2 100.0  90.0  2.3
# """

# obj = TOFPeakL.from_cif(s_cont)
# print(obj, end="\n\n")



# -*- coding: utf-8 -*-
"""
Calculations structure factor based on items and loops
"""
import numpy

from cryspy.C_item_loop_classes.cl_2_space_group import SpaceGroup
from cryspy.C_item_loop_classes.cl_1_cell import Cell
from cryspy.C_item_loop_classes.cl_1_atom_site import AtomSiteL
from cryspy.C_item_loop_classes.cl_1_atom_site_aniso import AtomSiteAnisoL

def calc_b_iso_beta(atom_site: AtomSiteL, atom_site_aniso: AtomSiteAnisoL, cell: Cell):
    """
    Calculate b_iso and beta_ij based on atom_site and atom_sites.
    For each atom defined in atom_site.
    """
    a_s = atom_site
    a_s_a = atom_site_aniso

    l_b_iso, l_beta = [], []
    coeff = float(8.*numpy.pi**2)

    for item_a_s in a_s.items:
        label_atom = item_a_s.label
        try:
            adp_type = item_a_s.adp_type
        except AttributeError:
            adp_type = None
        b_iso = 0.
        beta = (0., 0., 0., 0., 0., 0.)
        if adp_type == "Uiso":
            u_iso = float(item_a_s.u_iso_or_equiv)
            b_iso = float(8.*numpy.pi**2*u_iso)
        elif adp_type == "Biso":
            b_iso = float(item_a_s.b_iso_or_equiv)
        elif adp_type == "Uovl":
            # FIXME: correct it
            u_iso = float(item_a_s.u_iso_or_equiv)
            b_iso = coeff*u_iso
        elif adp_type == "Umpe":
            # FIXME: correct it
            u_iso = float(item_a_s.u_iso_or_equiv)
            b_iso = float(8.*numpy.pi**2*u_iso)
        elif adp_type == "Uani":
            if a_s_a is not None:
                item_a_s_a = a_s_a[label_atom]
                beta = item_a_s_a.calc_beta(cell)
        elif adp_type == "Bovl":
            # FIXME: correct it
            b_iso = float(item_a_s.b_iso_or_equiv)
        elif adp_type == "Bani":
            if a_s_a is not None:
                item_a_s_a = a_s_a[label_atom]
                beta = (float(item_a_s_a.b_11), float(item_a_s_a.b_22),
                        float(item_a_s_a.b_33), float(item_a_s_a.b_12),
                        float(item_a_s_a.b_13), float(item_a_s_a.b_23))
        l_b_iso.append(b_iso)
        l_beta.append(beta)
    np_b_iso = numpy.array(l_b_iso, dtype=float)
    np_beta = numpy.array(l_beta, dtype=float).transpose()
    return np_b_iso, np_beta




"""Classes Exclude, ExcludeL."""
from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class Exclude(ItemN):
    """
    Excluding region in data refinement.

    Attributes
    ----------
        - ttheta_min, ttheta_max, id, phi_min, phi_max,
          time_min, time_max (optional)

    Attributes time_min, time_max are used for TOF calculations
    """

    ATTR_MANDATORY_NAMES = ()
    ATTR_MANDATORY_TYPES = ()
    ATTR_MANDATORY_CIF = ()

    ATTR_OPTIONAL_NAMES = ("id", "ttheta_min", "ttheta_max", "phi_min",
                           "phi_max", "time_min", "time_max", "gamma_min", "gamma_max", "nu_min", "nu_max")
    ATTR_OPTIONAL_TYPES = (str, float, float, float, float, float, float, float, float, float, float)
    ATTR_OPTIONAL_CIF = ("id", "2theta_min", "2theta_max", "phi_min",
                         "phi_max", "time_min", "time_max", "gamma_min", "gamma_max", "nu_min", "nu_max")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "exclude"

    def __init__(self, **kwargs) -> NoReturn:
        super(Exclude, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"ttheta_min": 0., "ttheta_max": 0., "phi_min": -90.,
                 "phi_max": -90., "time_low": 0., "time_high": 0.}

        # defined for ani integer and float parameters
        D_MAX = {"ttheta_min": 180., "ttheta_max": 180., "phi_min": 90.,
                 "phi_max": 90.}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class ExcludeL(LoopN):
    """    Excluding regions in data refinement.
    """

    ITEM_CLASS = Exclude
    ATTR_INDEX = "id"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(ExcludeL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

# s_cont = """
#  loop_
#  _exclude_id
#  _exclude_ttheta_min
#  _exclude_ttheta_max
#   1   4.0  12.0
#   2  30.0  45.0
#   3  58.0  63.0
# """

# obj = ExcludeL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["2"], end="\n\n")
# print(obj.ttheta_min, end="\n\n")



"""Setup and SetupL classes."""
from typing import NoReturn
import numpy

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class Setup(ItemN):
    """Experimental diffraction setup (constant wavelength).

    Attributes
    ----------
        - wavelength (mandatory) (in Angstrems)
        - field (optional) (in Tesla)
        - radiation (optional) (neutrons by default, or X-rays)
        - offset_ttheta (optional for powder 1d and 2d) (in degrees)
        - offset_phi (optional for powder 2d) (in degrees)
        - ratio_lambdaover2 (optional, for single diffraction)
        - k (0. for neutrons, 0.5 for characteristic X-ray, 0.1 for synchrotron radiation)
        - cthm (cos**2 (2 theta_M)) (for calculation of Lorentrz polarization factor)
    """

    ATTR_MANDATORY_NAMES = ()
    ATTR_MANDATORY_TYPES = ()
    ATTR_MANDATORY_CIF = ()

    ATTR_OPTIONAL_NAMES = ("wavelength", "field", "offset_ttheta", "offset_phi", "offset_gamma", "offset_nu",
                           "ratio_lambdaover2", "radiation", "k", "cthm", "temperature")
    ATTR_OPTIONAL_TYPES = (float, float, float, float, float, float, float, str, float, float, float)
    ATTR_OPTIONAL_CIF = ("wavelength", "field", "offset_2theta", "offset_phi", "offset_gamma", "offset_nu",
                         "ratio_lambda/2", "radiation", "K", "cthm", "temperature")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("wavelength", "offset_ttheta", "offset_phi", "offset_gamma", "offset_nu",
                "ratio_lambdaover2")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {'wavelength': "{:.4f}", 'field': "{:.2f}",
                 'offset_ttheta': "{:.3f}", 'offset_phi': "{:.3f}",
                 'offset_gamma': "{:.3f}", 'offset_nu': "{:.3f}",
                 "ratio_lambdaover2": "{:.3f}", "k": "{:.1f}", "cthm": "{:.5f}",
                 'temperature': "{:.2f}",}

    # constraints on the parameters
    D_CONSTRAINTS = {"radiation": ["neutrons", "X-rays"]}

    # default values for the parameters
    D_DEFAULT = {"offset_2theta": 0., "radiation": "neutrons", "k":0., "cthm": 0.91}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "setup"

    def __init__(self, **kwargs) -> NoReturn:
        super(Setup, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"wavelength": 0., "ratio_lambdaover2": 0.}

        # defined for ani integer and float parameters
        D_MAX = {"ratio_lambdaover2": 1.}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def get_dictionary(self):
        res = {}
        if self.is_attribute("field"):
            res["magnetic_field"] = numpy.array([self.field], dtype=float)
        if self.is_attribute("temperature"):
            res["temperature"] = numpy.array([self.temperature], dtype=float)
        if self.is_attribute("wavelength"):
            res["wavelength"] = numpy.array([self.wavelength], dtype=float)
            res["flags_wavelength"] = numpy.array([self.wavelength_refinement], dtype=bool)
        if self.is_attribute("offset_ttheta"):
            if self.offset_ttheta is not None:
                res["offset_ttheta"] = numpy.array([self.offset_ttheta * numpy.pi/180.], dtype=float)
                res["flags_offset_ttheta"] = numpy.array([self.offset_ttheta_refinement], dtype=bool)
        if self.is_attribute("offset_gamma"):
            if self.offset_gamma is not None:
                res["offset_gamma"] = numpy.array([self.offset_gamma * numpy.pi/180.], dtype=float)
                res["flags_offset_gamma"] = numpy.array([self.offset_gamma_refinement], dtype=bool)
        if self.is_attribute("offset_nu"):
            if self.offset_nu is not None:
                res["offset_nu"] = numpy.array([self.offset_nu * numpy.pi/180.], dtype=float)
                res["flags_offset_nu"] = numpy.array([self.offset_nu_refinement], dtype=bool)
        if self.is_attribute("radiation"):
            res["radiation"] = numpy.array([self.radiation], dtype=str)
        if self.is_attribute("k"):
            res["k"] = numpy.array([self.k], dtype=float)
        if self.is_attribute("cthm"):
            res["cthm"] = numpy.array([self.cthm], dtype=float)
        if self.is_attribute("ratio_lambdaover2"):
            res["c_lambda2"] = numpy.array([self.ratio_lambdaover2], dtype=float)
            res["flags_c_lambda2"] = numpy.array([self.ratio_lambdaover2_refinement], dtype=bool)

        return res

class SetupL(LoopN):
    """Experimental diffraction setup (constant wavelength).
    """

    ITEM_CLASS = Setup
    ATTR_INDEX = None

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(SetupL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name



"""Space Group class."""

from typing import NoReturn
import numpy
from fractions import Fraction

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary
from cryspy.A_functions_base.function_2_space_group import \
    get_shift_by_centring_type, \
    get_symop_pcentr_multiplicity_letter_site_symmetry_coords_xyz_2, \
    get_type_hm, auto_choose_it_coordinate_system_code, \
    get_it_number_it_coordinate_system_codes_by_name_hm_extended,\
    get_it_number_by_name_hm_full, \
    get_it_number_by_name_hm_short, \
    get_it_number_by_name_hall, \
    get_it_number_by_name_schoenflies, \
    get_it_coordinate_system_codes_by_it_number, \
    get_default_it_coordinate_system_code_by_it_number, \
    get_crystal_system_by_it_number, \
    get_name_hm_extended_by_it_number_it_coordinate_system_code, \
    get_centring_type_by_name_hm_extended, \
    get_bravais_type_by_centring_type_crystal_system, \
    get_name_hm_short_by_it_number, \
    get_lattice_type_by_name_hm_short, \
    get_name_hm_full_by_it_number, \
    get_name_hall_by_it_number, \
    get_centrosymmetry_by_name_hall, \
    get_name_schoenflies_by_it_number, \
    get_laue_class_by_name_schoenflies, \
    get_point_group_hm_short_by_name_schoenflies, \
    get_generators_by_point_group_hm, \
    get_patterson_name_hm_by_lattice_type_laue_class, \
    ACCESIBLE_IT_NUMBER, ACCESIBLE_BRAVAIS_TYPE, \
    ACCESIBLE_IT_COORDINATE_SYSTEM_CODE, ACCESIBLE_LAUE_CLASS, \
    ACCESIBLE_CENTRING_TYPE, ACCESIBLE_CRYSTAL_SYSTEM, \
    ACCESIBLE_NAME_HM_SHORT, ACCESIBLE_NAME_HM_FULL, \
    ACCESIBLE_NAME_HM_EXTENDED, ACCESIBLE_NAME_SCHOENFLIES, \
    ACCESIBLE_NAME_HALL_SHORT, ACCESIBLE_REFERENCE_SETTING

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from cryspy.C_item_loop_classes.cl_1_space_group_symop import \
    SpaceGroupSymop, SpaceGroupSymopL

from cryspy.C_item_loop_classes.cl_1_space_group_wyckoff import \
    SpaceGroupWyckoff, SpaceGroupWyckoffL


class SpaceGroup(ItemN):
    """Space group description.

    Contains all the data items that refer to the space group as a
    whole, such as its name, Laue group etc.

    Space-group types are identified by their number as listed in
    International Tables for Crystallography Volume A, or by their
    Schoenflies symbol. Specific settings of the space groups can
    be identified by their Hall symbol, by specifying their
    symmetry operations or generators, or by giving the
    transformation that relates the specific setting to the
    reference setting based on International Tables Volume A and
    stored in this dictionary.

    The commonly used Hermann-Mauguin symbol determines the
    space-group type uniquely but several different Hermann-Mauguin
    symbols may refer to the same space-group type. A
    Hermann-Mauguin symbol contains information on the choice of
    the basis, but not on the choice of origin.

    """

    ATTR_MANDATORY_NAMES = ()
    ATTR_MANDATORY_TYPES = ()
    ATTR_MANDATORY_CIF = ()

    ATTR_OPTIONAL_NAMES = (
        "id", "name_hm_alt", "name_hm_alt_description", "it_number",
        "name_hm_ref", "it_coordinate_system_code", "name_hm_full",
        "name_hall", "name_schoenflies", "point_group_hm", "laue_class",
        "patterson_name_hm", "centring_type", "bravais_type", "crystal_system",
        "reference_setting", "transform_pp_abc", "transform_qq_xyz")
    ATTR_OPTIONAL_TYPES = (str, str, str, int, str, str, str, str, str, str,
                           str, str, str, str, str, str, str, str)
    ATTR_OPTIONAL_CIF = (
        "id", "name_H-M_alt", "name_H-M_alt_description", "IT_number",
        "name_H-M_ref", "IT_coordinate_system_code", "name_H-M_full",
        "name_Hall", "name_Schoenflies", "point_group_H-M", "Laue_class",
        "Patterson_name_H-M", "centring_type", "Bravais_type",
        "crystal_system", "reference_setting", "transform_pp_abc",
        "transform_qq_xyz")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ("centrosymmetry", "pcentr", "reduced_space_group_symop",
                      "full_space_group_symop", "space_group_wyckoff", "shift")
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    ACCESIBLE_NAME_HM_ALT = frozenset(
        set(ACCESIBLE_NAME_HM_SHORT) | set(ACCESIBLE_NAME_HM_FULL) |
        set(ACCESIBLE_NAME_HM_EXTENDED))

    # constraints on the parameters
    D_CONSTRAINTS = {
        "it_number": ACCESIBLE_IT_NUMBER,
        "bravais_type": ACCESIBLE_BRAVAIS_TYPE,
        "it_coordinate_system_code": ACCESIBLE_IT_COORDINATE_SYSTEM_CODE,
        "laue_class": ACCESIBLE_LAUE_CLASS,
        "centring_type": ACCESIBLE_CENTRING_TYPE,
        "crystal_system": ACCESIBLE_CRYSTAL_SYSTEM,
        "name_hm_alt": ACCESIBLE_NAME_HM_ALT,
        "name_hm_ref": ACCESIBLE_NAME_HM_SHORT,
        "name_schoenflies": ACCESIBLE_NAME_SCHOENFLIES,
        "name_hall": ACCESIBLE_NAME_HALL_SHORT,
        "reference_setting": ACCESIBLE_REFERENCE_SETTING,
        "name_hm_full": ACCESIBLE_NAME_HM_FULL}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "space_group"

    def __init__(self, **kwargs) -> NoReturn:
        super(SpaceGroup, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        flag_form_object = False
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)
            if key in ("it_number", "name_hm_alt", "name_hm_ref",
                       "name_schoenflies", "name_hall", "name_hm_full"):
                flag_form_object = True

        if flag_form_object:
            self.form_object()

    def form_object(self) -> NoReturn:
        """Form object."""
        self.form_object_by_it_number_it_coordinate_system_code()
        self.__dict__["shift"] = get_shift_by_centring_type(self.centring_type)

    def define_it_number_and_it_coordinate_system_code(self):
        """Define it number and it coordinate system code."""
        flag = True
        it_coordinate_system_code = None
        if self.is_attribute("it_coordinate_system_code"):
            it_coordinate_system_code = self.it_coordinate_system_code
        
        if (self.is_attribute("it_number") &
                self.is_attribute("it_coordinate_system_code")):
            it_number = self.it_number
            it_coordinate_system_code = self.it_coordinate_system_code
        elif ((self.is_attribute("name_hm_alt")) &
              (it_coordinate_system_code is not None)):
            t_res = get_type_hm(self.name_hm_alt)
            if "full" in t_res:
                it_number = get_it_number_by_name_hm_full(self.name_hm_alt)
            elif "short" in t_res:
                it_number = get_it_number_by_name_hm_short(self.name_hm_alt)
            elif "extended" in t_res:
                it_number, it_coordinate_system_codes = \
                    get_it_number_it_coordinate_system_codes_by_name_hm_extended(
                        self.name_hm_alt)
                if (not(it_coordinate_system_code in
                        it_coordinate_system_codes)):
                    it_coordinate_system_code = \
                        auto_choose_it_coordinate_system_code(
                            it_number, it_coordinate_system_codes)
        elif self.is_attribute("name_hm_alt"):
            t_res = get_type_hm(self.name_hm_alt)
            if "extended" in t_res:
                it_number, it_coordinate_system_codes = \
                    get_it_number_it_coordinate_system_codes_by_name_hm_extended(
                        self.name_hm_alt)
                if (not(it_coordinate_system_code in
                        it_coordinate_system_codes)):
                    it_coordinate_system_code = \
                        auto_choose_it_coordinate_system_code(
                            it_number, it_coordinate_system_codes)
            elif "full" in t_res:
                it_number = get_it_number_by_name_hm_full(self.name_hm_alt)
            elif "short" in t_res:
                it_number = get_it_number_by_name_hm_short(self.name_hm_alt)
        elif self.is_attribute("name_hm_ref"):
            it_number = get_it_number_by_name_hm_short(self.name_hm_ref)
        elif self.is_attribute("name_hall"):
            it_number = get_it_number_by_name_hall(self.name_hall)
        elif self.is_attribute("it_number"):
            it_number = self.it_number
            # it_coordinate_system_code = self.it_coordinate_system_code
        elif self.is_attribute("name_schoenflies"):
            it_number = get_it_number_by_name_schoenflies(
                self.name_schoenflies)
        else:
            flag = False
        if flag:
            it_coordinate_system_codes = \
                get_it_coordinate_system_codes_by_it_number(it_number)
            self.__dict__["it_number"] = it_number
            if (not(it_coordinate_system_code in it_coordinate_system_codes)):
                it_coordinate_system_code = \
                    auto_choose_it_coordinate_system_code(
                        it_number, it_coordinate_system_codes)
            self.__dict__["it_coordinate_system_code"] = \
                it_coordinate_system_code
        return flag

    def form_object_by_it_number_it_coordinate_system_code(self):
        """Form object by it number it coordinate system code.

        TODO: Solve the problem with centring_type for hexagonal systems
              (Ex.: 166 space group).
        """
        bravais_type, laue_class, patterson_name_hm, centring_type, \
            crystal_system = None, None, None, None, None
        name_hm_extended, name_hm_full, name_hm_short = None, None, None
        name_hall, name_schoenflies, point_group_hm = None, None, None
        pcentr = None
        reduced_space_group_symop, full_space_group_symop = None, None
        space_group_wyckoff = None

        lattice_type = None
        generators = ()
        self.define_it_number_and_it_coordinate_system_code()
        try:
            it_number = self.it_number
        except AttributeError:
            it_number = None
        try:
            it_c_s_c = self.it_coordinate_system_code
        except AttributeError:
            it_c_s_c = None
        crystal_system = get_crystal_system_by_it_number(it_number)
        name_hm_extended = \
            get_name_hm_extended_by_it_number_it_coordinate_system_code(
                it_number, it_c_s_c)
        if (name_hm_extended is not None):
            centring_type = get_centring_type_by_name_hm_extended(
                name_hm_extended)
        name_hm_short = get_name_hm_short_by_it_number(it_number)
        if (name_hm_short is not None):
            lattice_type = get_lattice_type_by_name_hm_short(name_hm_short)
            if centring_type is None:
                centring_type = get_centring_type_by_name_hm_extended(
                    name_hm_short)
        name_hm_full = get_name_hm_full_by_it_number(it_number)
        if ((name_hm_full is not None) & (centring_type is None)):
            centring_type = get_centring_type_by_name_hm_extended(name_hm_full)
        if ((centring_type is not None) & (crystal_system is not None)):
            bravais_type = get_bravais_type_by_centring_type_crystal_system(
                centring_type, crystal_system)
        name_hall = get_name_hall_by_it_number(it_number)
        if name_hall is not None:
            centrosymmetry = get_centrosymmetry_by_name_hall(name_hall)
        name_schoenflies = get_name_schoenflies_by_it_number(it_number)
        if name_schoenflies is not None:
            laue_class = get_laue_class_by_name_schoenflies(name_schoenflies)
            point_group_hm = get_point_group_hm_short_by_name_schoenflies(
                name_schoenflies)
            if point_group_hm is not None:
                generators = get_generators_by_point_group_hm(point_group_hm)
            if ((lattice_type is not None) & (laue_class is not None)):
                patterson_name_hm = \
                    get_patterson_name_hm_by_lattice_type_laue_class(
                        lattice_type, laue_class)
        if ((centring_type is not None) and (it_c_s_c is not None)):
            if centring_type.startswith("R") and it_c_s_c.startswith("r"):
                centring_type = "P"

        pcentr, reduced_space_group_symop, full_space_group_symop, \
            space_group_wyckoff = \
            self.form_reduced_full_space_group_symop_space_group_wyckoff(
                it_number, it_c_s_c, centrosymmetry, centring_type)

        if pcentr is not None:
            self.__dict__["pcentr"] = pcentr
        if reduced_space_group_symop is not None:
            self.__dict__["reduced_space_group_symop"] = \
                reduced_space_group_symop
        if full_space_group_symop is not None:
            self.__dict__["full_space_group_symop"] = full_space_group_symop
        if space_group_wyckoff is not None:
            self.__dict__["space_group_wyckoff"] = space_group_wyckoff

        if centrosymmetry is not None:
            self.__dict__["centrosymmetry"] = centrosymmetry
        # if name_hm_extended is not None:
        #     self.__dict__["name_hm_alt"] = name_hm_extended
        #     self.__dict__["name_hm_alt_description"] = \
        #         "The extended Hermann-Mauguin space-group symbol."
        if name_hm_full is not None:
            self.__dict__["name_hm_full"] = name_hm_full
        if name_hm_short is not None:
            self.__dict__["name_hm_ref"] = name_hm_short
            self.__dict__["name_hm_alt"] = name_hm_short
            self.__dict__["name_hm_alt_description"] = \
                "The short Hermann-Mauguin space-group symbol."
        if name_hall is not None:
            self.__dict__["name_hall"] = name_hall
        if name_schoenflies is not None:
            self.__dict__["name_schoenflies"] = name_schoenflies

        if point_group_hm is not None:
            self.__dict__["point_group_hm"] = point_group_hm
        if laue_class is not None:
            self.__dict__["laue_class"] = laue_class
        if patterson_name_hm is not None:
            self.__dict__["patterson_name_hm"] = patterson_name_hm
        if centring_type is not None:
            self.__dict__["centring_type"] = centring_type
        if bravais_type is not None:
            self.__dict__["bravais_type"] = bravais_type
        if crystal_system is not None:
            self.__dict__["crystal_system"] = crystal_system

    def form_reduced_full_space_group_symop_space_group_wyckoff(
            self, it_number: int, it_coordinate_system_code: str,
            centrosymmetry: bool, centring_type: str):
        """Form reduced full space group symop space group Wyckoff."""
        pcentr, reduced_space_group_symop, \
            full_space_group_symop = None, None, None

        symop, pcentr, _multiplicity, _letter, _site_symmetry, _l_coords_xyz_2\
            = get_symop_pcentr_multiplicity_letter_site_symmetry_coords_xyz_2(
                it_number, it_coordinate_system_code)

        item_reduced = []
        for _i_symop, _symop in enumerate(symop):
            _item = SpaceGroupSymop(id=f"{_i_symop+1:}", operation_xyz=_symop,
                                    operation_description="reduced",
                                    sg_id=it_number)
            item_reduced.append(_item)
        reduced_space_group_symop = SpaceGroupSymopL()
        reduced_space_group_symop.items = item_reduced
        
        item_full = []
        for _item in item_reduced:
            _symop = _item.operation_xyz
            _item_full = SpaceGroupSymop(operation_xyz=_symop)
            item_full.append(_item_full)
        if centrosymmetry:
            item_add = []
            for _item in item_full:
                _item_add = _item.get_symop_inversed(pcentr)
                item_add.append(_item_add)
            item_full.extend(item_add)
        item_new = []
        for _item in item_full:
            _items_new = _item.get_symops_by_centring_type(centring_type)
            item_new.extend(_items_new)
        for _i, _item in enumerate(item_new):
            _item.id = f"{_i+1:}"
            _item.sg_id = f"{it_number:}"
            _item.operation_description = "full"
        full_space_group_symop = SpaceGroupSymopL()
        full_space_group_symop.items = item_new

        _i_numb = 0
        item = []
        for _1, _2, _3, _4 in zip(_multiplicity, _letter, _site_symmetry,
                                  _l_coords_xyz_2):
            _i_numb += 1
            _item = SpaceGroupWyckoff(id=f"{_i_numb:}", multiplicity=_1,
                                      coord_xyz=_4[0], letter=_2,
                                      site_symmetry=_3)
            _item.it_coord_xyz = _4
            _item.centring_type = centring_type
            _item.form_object()
            item.append(_item)
        space_group_wyckoff = SpaceGroupWyckoffL()
        space_group_wyckoff.items = item
        return pcentr, reduced_space_group_symop, full_space_group_symop, \
            space_group_wyckoff

    @staticmethod
    def get_it_coordinate_system_codes_by_it_number(it_number: int) -> str:
        """Get it coordinate system codes by it number."""
        return get_it_coordinate_system_codes_by_it_number(it_number)

    @staticmethod
    def get_default_it_coordinate_system_code_by_it_number(it_number: int) \
            -> str:
        """Get default it coordinate system code by it number."""
        return get_default_it_coordinate_system_code_by_it_number(it_number)

    def calc_hkl_equiv(self, index_h, index_k, index_l):
        """Gequivalent reflections of hkl and its multiplicity."""
        r_11 = numpy.array(self.reduced_space_group_symop.r_11, dtype=Fraction)
        r_12 = numpy.array(self.reduced_space_group_symop.r_12, dtype=Fraction)
        r_13 = numpy.array(self.reduced_space_group_symop.r_13, dtype=Fraction)
        r_21 = numpy.array(self.reduced_space_group_symop.r_21, dtype=Fraction)
        r_22 = numpy.array(self.reduced_space_group_symop.r_22, dtype=Fraction)
        r_23 = numpy.array(self.reduced_space_group_symop.r_23, dtype=Fraction)
        r_31 = numpy.array(self.reduced_space_group_symop.r_31, dtype=Fraction)
        r_32 = numpy.array(self.reduced_space_group_symop.r_32, dtype=Fraction)
        r_33 = numpy.array(self.reduced_space_group_symop.r_33, dtype=Fraction)

        h_s = r_11*index_h + r_21*index_k + r_31*index_l
        k_s = r_12*index_h + r_22*index_k + r_32*index_l
        l_s = r_13*index_h + r_23*index_k + r_33*index_l

        hkl_s = numpy.vstack([h_s, k_s, l_s])
        hkl_s = numpy.hstack([hkl_s, -1*hkl_s])
        hkl_s = hkl_s.astype(int)
        hkl_s_un = numpy.unique(hkl_s, axis=1)
        multiplicity = int(round(hkl_s.shape[1]*1./hkl_s_un.shape[1]))
        h_s, k_s, l_s = hkl_s_un[0, :], hkl_s_un[1, :], hkl_s_un[2, :]
        return h_s, k_s, l_s, multiplicity

    def calc_xyz_mult(self, x, y, z):
        """Give unique x,y,z elements and calculate multiplicity.

        It's done for given x,y,z fract.
        """
        wyckoff = self.space_group_wyckoff.get_wyckoff_for_fract(x, y, z)
        np_r = numpy.array(wyckoff.full_r, dtype=float)
        np_b = numpy.array(wyckoff.full_b, dtype=float)
        x_s = (np_r[:, 0, 0]*x + np_r[:, 0, 1]*y + np_r[:, 0, 2]*z +
               np_b[:, 0]) % 1
        y_s = (np_r[:, 1, 0]*x + np_r[:, 1, 1]*y + np_r[:, 1, 2]*z +
               np_b[:, 1]) % 1
        z_s = (np_r[:, 2, 0]*x + np_r[:, 2, 1]*y + np_r[:, 2, 2]*z +
               np_b[:, 2]) % 1

        # l_shift = self.shift
        # l_x, l_y, l_z = [], [], []
        # for _shift in l_shift:
        #    l_x.extend(numpy.mod(x_s+_shift[0],1))
        #    l_y.extend(numpy.mod(y_s+_shift[1],1))
        #    l_z.extend(numpy.mod(z_s+_shift[2],1))
        multiplicity = wyckoff.multiplicity
        # x_out = numpy.array(l_x, dtype=float)
        # y_out = numpy.array(l_y, dtype=float)
        # z_out = numpy.array(l_z, dtype=float)

        return x_s, y_s, z_s, multiplicity

    def calc_symop_for_xyz(self, x_in, y_in, z_in):
        """Calculate symop for xyz."""
        x, y, z = x_in % 1., y_in % 1., z_in % 1.

        symop = self.full_space_group_symop
        e_11 = numpy.array(symop.r_11, dtype=float)
        e_12 = numpy.array(symop.r_12, dtype=float)
        e_13 = numpy.array(symop.r_13, dtype=float)
        e_1 = numpy.array(symop.b_1, dtype=float)

        e_21 = numpy.array(symop.r_21, dtype=float)
        e_22 = numpy.array(symop.r_22, dtype=float)
        e_23 = numpy.array(symop.r_23, dtype=float)
        e_2 = numpy.array(symop.b_2, dtype=float)

        e_31 = numpy.array(symop.r_31, dtype=float)
        e_32 = numpy.array(symop.r_32, dtype=float)
        e_33 = numpy.array(symop.r_33, dtype=float)
        e_3 = numpy.array(symop.b_3, dtype=float)

        x_s = numpy.round(numpy.mod(e_11*x + e_12*y + e_13*z + e_1, 1), decimals=5)
        y_s = numpy.round(numpy.mod(e_21*x + e_22*y + e_23*z + e_2, 1), decimals=5)
        z_s = numpy.round(numpy.mod(e_31*x + e_32*y + e_33*z + e_3, 1), decimals=5)

        xyz_s = numpy.vstack([x_s, y_s, z_s])

        xyz_s_un, unique_inverse = numpy.unique(xyz_s, return_inverse=True,
                                                axis=1)
        x_s, y_s, z_s = xyz_s_un[0, :], xyz_s_un[1, :], xyz_s_un[2, :]
        ind = (numpy.where((x-x_s)**2+(y-y_s)**2+(z-z_s)**2 < 0.00001))[0][0]

        flag = unique_inverse == ind

        item_out = [_item for _item, _flag in zip(symop.items, flag) if _flag]
        symop_out = SpaceGroupSymopL()
        symop_out.items = item_out
        return symop_out

    def calc_el_symm_for_xyz(self, x_in, y_in, z_in):
        """FIXME: should be deleted."""
        x, y, z = x_in % 1., y_in % 1., z_in % 1.

        symop = self.full_space_group_symop
        e_11 = numpy.array(symop.r_11, dtype=float)
        e_12 = numpy.array(symop.r_12, dtype=float)
        e_13 = numpy.array(symop.r_13, dtype=float)
        e_1 = numpy.array(symop.b_1, dtype=float)

        e_21 = numpy.array(symop.r_21, dtype=float)
        e_22 = numpy.array(symop.r_22, dtype=float)
        e_23 = numpy.array(symop.r_23, dtype=float)
        e_2 = numpy.array(symop.b_2, dtype=float)

        e_31 = numpy.array(symop.r_31, dtype=float)
        e_32 = numpy.array(symop.r_32, dtype=float)
        e_33 = numpy.array(symop.r_33, dtype=float)
        e_3 = numpy.array(symop.b_3, dtype=float)

        x_s = numpy.round(numpy.mod(e_11*x + e_12*y + e_13*z + e_1, 1), decimals=5)
        y_s = numpy.round(numpy.mod(e_21*x + e_22*y + e_23*z + e_2, 1), decimals=5)
        z_s = numpy.round(numpy.mod(e_31*x + e_32*y + e_33*z + e_3, 1), decimals=5)

        xyz_s = numpy.vstack([x_s, y_s, z_s])

        xyz_s_un, unique_inverse = numpy.unique(xyz_s, return_inverse=True,
                                                axis=1)
        # Output for unique_inverse in version of numpy 2.0.0 and 1.23.5 is different. 
        # This line is to unified two versions
        unique_inverse = unique_inverse.flatten()
        x_s, y_s, z_s = xyz_s_un[0, :], xyz_s_un[1, :], xyz_s_un[2, :]
        ind = (numpy.where((x-x_s)**2+(y-y_s)**2+(z-z_s)**2 < 0.00001))[0][0]
        flag = unique_inverse == ind
        o_11, o_12, o_13 = e_11[flag], e_12[flag], e_13[flag]
        o_21, o_22, o_23 = e_21[flag], e_22[flag], e_23[flag]
        o_31, o_32, o_33 = e_31[flag], e_32[flag], e_33[flag]
        o_1, o_2, o_3 = e_1[flag], e_2[flag], e_3[flag]
        return o_11, o_12, o_13, o_21, o_22, o_23, o_31, o_32, o_33, o_1, \
            o_2, o_3

    def calc_f_hkl_by_f_hkl_as(self, index_h, index_k, index_l, f_hkl_as):
        """Calculate the structure factor.

        The structure factor in asymmetric unit cell are given.
        """
        np_h = numpy.array(index_h, dtype=float)
        np_k = numpy.array(index_k, dtype=float)
        np_l = numpy.array(index_l, dtype=float)
        shift = numpy.array(self.shift, dtype=float)
        centr = self.centrosymmetry

        orig_x, orig_y, orig_z = shift[:, 0], shift[:, 1], shift[:, 2]

        # orig_x = [hh[0] for hh in shift]
        # orig_y = [hh[1] for hh in shift]
        # orig_z = [hh[2] for hh in shift]

        np_h_2d, np_orig_x_2d = numpy.meshgrid(index_h, orig_x, indexing="ij")
        np_k_2d, np_orig_y_2d = numpy.meshgrid(index_k, orig_y, indexing="ij")
        np_l_2d, np_orig_z_2d = numpy.meshgrid(index_l, orig_z, indexing="ij")

        hh = (2*numpy.pi*1j*(np_h_2d*np_orig_x_2d + np_k_2d*np_orig_y_2d +
                             np_l_2d*np_orig_z_2d)).astype(complex)
        np_orig_as = numpy.exp(hh)
        _hh = np_orig_as.sum(axis=1)
        if len(f_hkl_as.shape) == 2:
            _hh = _hh[:, numpy.newaxis]
        f_hkl_1 = f_hkl_as*_hh*1./len(shift)

        if (centr):
            p_centr = numpy.array(self.pcentr, dtype=float)
            hh = (2. * 2. * numpy.pi * 1j * (
                np_h*p_centr[0] + np_k*p_centr[1] + np_l*p_centr[2])
                ).astype(complex)
            if len(f_hkl_as.shape) == 2:
                hh = hh[:, numpy.newaxis]
            f_hkl = 0.5*(f_hkl_1+f_hkl_1.conjugate()*numpy.exp(hh))
        else:
            f_hkl = f_hkl_1
        return f_hkl

    def calc_asymmetric_cell(self, n_a, n_b, n_c):
        """Give the numbers in asymmetric cell.

         :n_a: the number of points along a axis
         :n_b: the numper of points along b axis
         :n_c: the numper of points along c axis

        na, n_b, nc should be divided on 24: 8 and 3

        :output: - l_coord is a list of coordinates in asymmetric cell
                   (frac_x = n_x/n_a and so on)
                 - l_symm contains a list of symmetry given as
                   (n_symm, centr, n_orig)
        """
        n_a_new = int(round(n_a/24))*24
        # n_b_new = int(round(n_b/24))*24
        # n_c_new = int(round(n_c/24))*24

        # l_el_symm = self.el_symm
        # f_centr = self.centr
        # p_centr = self.p_centr
        # l_orig = self.orig
        l_coord = []

        spgr_choice = self.spgr_choice
        # spgr_name = self.spgr_name
        spgr_number = self.spgr_number

        if (spgr_number == 227) & (spgr_choice == "2"):
            n_a_new = int(round(n_a/8))*8
            for n_x in range(-n_a_new//8, 3*n_a_new//8+1):
                for n_y in range(-n_a_new//8, 0+1):
                    for n_z in range(-n_a_new//4, 0+1):
                        cond_1 = (n_y < min([n_a_new//4-n_x, n_x]))
                        cond_2 = (n_z >= -n_y-n_a_new//4)
                        cond_3 = (n_z <= n_y)
                        if (cond_1 & cond_2 & cond_3):
                            coord_x = float(n_x)/float(n_a_new)
                            coord_y = float(n_y)/float(n_a_new)
                            coord_z = float(n_z)/float(n_a_new)
                            l_coord.append((coord_x, coord_y, coord_z))
        return l_coord

    def calc_rotated_matrix_for_position(self, m_chi, x, y, z):
        """Calc rotated matrix for position."""
        symop = self.full_space_group_symop
        e_11 = numpy.array(symop.r_11, dtype=float)
        e_12 = numpy.array(symop.r_12, dtype=float)
        e_13 = numpy.array(symop.r_13, dtype=float)
        e_1 = numpy.array(symop.b_1, dtype=float)

        e_21 = numpy.array(symop.r_21, dtype=float)
        e_22 = numpy.array(symop.r_22, dtype=float)
        e_23 = numpy.array(symop.r_23, dtype=float)
        e_2 = numpy.array(symop.b_2, dtype=float)

        e_31 = numpy.array(symop.r_31, dtype=float)
        e_32 = numpy.array(symop.r_32, dtype=float)
        e_33 = numpy.array(symop.r_33, dtype=float)
        e_3 = numpy.array(symop.b_3, dtype=float)

        x_s = numpy.round(numpy.mod(e_11*x + e_12*y + e_13*z + e_1, 1), decimals=5)
        y_s = numpy.round(numpy.mod(e_21*x + e_22*y + e_23*z + e_2, 1), decimals=5)
        z_s = numpy.round(numpy.mod(e_31*x + e_32*y + e_33*z + e_3, 1), decimals=5)
        # o_11, o_12, o_13, o_21, o_22, o_23, o_31, o_32, o_33, o_1, o_2, o_3 \
        #    = self.calc_el_symm_for_xyz(x, y, z)
        # np_x, np_y, np_z, mult = self.calc_xyz_mult(x, y, z)

        l_ind, l_xyz = [], []
        _ind = 0
        for _x, _y, _z in zip(x_s, y_s, z_s):
            if (_x, _y, _z) not in l_xyz:
                l_ind.append(_ind)
                l_xyz.append((_x, _y, _z))
            _ind += 1
        l_res = []
        for _ind, _xyz in zip(l_ind, l_xyz):
            _11, _12, _13 = e_11[_ind], e_12[_ind], e_13[_ind]
            _21, _22, _23 = e_21[_ind], e_22[_ind], e_23[_ind]
            _31, _32, _33 = e_31[_ind], e_32[_ind], e_33[_ind]
            # _1, _2, _3 = e_1[_ind], e_2[_ind], e_3[_ind]
            matrix_r = numpy.array([[_11, _12, _13], [_21, _22, _23],
                                    [_31, _32, _33]], dtype=float)
            matrix_rt = matrix_r.transpose()
            r_chi = numpy.matmul(matrix_r, m_chi)
            matrix_chi_rot = numpy.matmul(r_chi, matrix_rt)
            l_res.append((_xyz, matrix_chi_rot))
        return l_res

    def report(self) -> str:
        return self.report_space_group()

    def report_space_group(self) -> str:
        """Make a report about space group in string format."""
        if self.is_defined:
            self.form_object
        else:
            return ""
        ls_out = []
        ls_out.append("# Space group")

        width_left, width_right = 30, 40

        if self.is_attribute("it_number"):
            ls_out.append("|IT_number: ".rjust(width_left) + "|" +
                          f"{self.it_number:}|".ljust(width_right))
        if self.is_attribute("name_hm_alt"):
            ls_out.append(
                "|Name H-M alt: ".rjust(width_left) + "|" +
                f"\"{self.name_hm_alt:}\"|".ljust(width_right))
        if self.is_attribute("name_hm_full"):
            ls_out.append(
                "|Name H-M full: ".rjust(width_left) + "|" +
                f"\"{self.name_hm_full:}\"|".ljust(width_right))
        if self.is_attribute("name_hm_ref"):
            ls_out.append(
                "|Name H-M ref: ".rjust(width_left) + "|" +
                f"\"{self.name_hm_ref:}\"|".ljust(width_right))
        if self.is_attribute("name_hall"):
            ls_out.append("|Name Hall short: ".rjust(width_left) + "|" +
                          f"\"{self.name_hall:}\"|".ljust(width_right))
        if self.is_attribute("name_schoenflies"):
            ls_out.append(
                "|Name Schoenflies: ".rjust(width_left) + "|" +
                f"\"{self.name_schoenflies:}\"|".ljust(width_right))
        if self.is_attribute("it_coordinate_system_code"):
            ls_out.append("|IT_coordinate_system_code: ".rjust(width_left) + "|" +
                          f"\"{self.it_coordinate_system_code:}\"|".ljust(
                              width_right))
        ls_out.append("")
        if self.is_attribute("point_group_hm"):
            ls_out.append(
                "|Point group H-M: ".rjust(width_left) + "|" +
                f"\"{self.point_group_hm:}\"|".ljust(width_right))
        if self.is_attribute("laue_class"):
            ls_out.append("|Laue class: ".rjust(width_left) + "|" +
                          f"\"{self.laue_class:}\"|".ljust(width_right))
        if self.is_attribute("patterson_name_hm"):
            ls_out.append(
                "|Patterson name H-M: ".rjust(width_left) + "|" +
                f"\"{self.patterson_name_hm:}\"|".ljust(width_right))
        if self.is_attribute("centring_type"):
            ls_out.append(
                "|Centring type: ".rjust(width_left) + "|" +
                f"\"{self.centring_type:}\"|".ljust(width_right))
        if self.is_attribute("bravais_type"):
            ls_out.append(
                "|Bravais type: ".rjust(width_left) + "|" +
                f"\"{self.bravais_type:}\"|".ljust(width_right))
        if self.is_attribute("crystal_system"):
            ls_out.append(
                "|Crystal system: ".rjust(width_left) + "|" +
                f"\"{self.crystal_system:}\"|".ljust(width_right))
        ls_out.append("")
        if self.is_attribute("centrosymmetry"):
            ls_out.append(
                "|Centrosymmetry: ".rjust(width_left) + "|" +
                f"{'Yes' if self.centrosymmetry else 'No':}|".ljust(
                    width_right))

        # if generators != (): print(
        #    f"Generators: ".rjust(width_left) +
        #    ", ".join([f"\"{_}\"" for _ in generators]).ljust(width_right))

        # if symop is not None:
        #    print("Symop: ")  # pcentr
        #    print_long_list([f"\"{_:}\"" for _ in symop])

        if self.is_attribute("space_group_wyckoff"):
            s_g_w = self.space_group_wyckoff
            ls_out.append("\n## Special positions:\n")
            ls_out.append("|Position | Multiplicity| Character| Group|")
            ls_out.append("|---------|-------------|----------|------|")
            for _1, _2, _3, _4 in zip(s_g_w.multiplicity, s_g_w.letter,
                                      s_g_w.site_symmetry, s_g_w.coord_xyz):
                ls_out.append(
                    f"|{_4.rjust(12):12}| {_1:3}| {_2.rjust(3):3}| \
{_3.rjust(7):7}|")

        if self.is_attribute("reduced_space_group_symop"):
            r_s_g_s = self.reduced_space_group_symop
            ls_out.append("\n## Reduced space group symop:\n")
            line = []
            for _i,  _1 in enumerate(r_s_g_s.operation_xyz):
                line.append(f" {_1.rjust(12):12}")
                if _i % 3 == 2:
                    ls_out.append("|"+"|".join(line)+"|")
                    line = []

        if self.is_attribute("shift"):
            shift = self.shift
            ls_out.append("\n## Shift:\n")
            ls_out.append("|along a | along b| along c|")
            ls_out.append("|--------|--------|--------|")
            for _pos in shift:
                ls_out.append(f"|{float(_pos[0]):9.5f}| {float(_pos[1]):9.5f}| \
{float(_pos[2]):9.5f}|")
        return "\n".join(ls_out)

    def to_cif(self, separator: str = "_", flag_all_attributes: bool = False,
               flag_minimal: bool = True) -> str:
        """Print information about object in string in STAR format.

        Arguments
        ---------
            prefix is a prefix in front of label of attribute
            separator is a separator between prefix and attribute ("_" or ".")
            flag if it's True the value "." will be printed for undefined
            attributes flag_minimal if it's True the minimal set of object will
            be printed

        Returns
        -------
            A string in STAR/CIF format
        """
        if flag_minimal:
            ls_out = []
            prefix = self.PREFIX
            attributes = ["name_hm_alt", "it_coordinate_system_code"]
            related_attributes = ["name_H-M_alt", "IT_coordinate_system_code"]
            for _attr, cif_attr in zip(attributes, related_attributes):
                if self.is_attribute(_attr):
                    _val = getattr(self, _attr)
                    s_val = str(_val)
                    if len(s_val.split(" ")) > 1:
                        ls_out.append(
                            f"_{prefix:}{separator:}{cif_attr:} \"{s_val:}\"")
                    else:
                        ls_out.append(
                            f"_{prefix:}{separator:}{cif_attr:} {s_val:}")
                else:
                    ls_out.append(f"_{prefix:}{separator:}{cif_attr:} .")
            s_out = "\n".join(ls_out)
        else:
            s_out = super(SpaceGroup, self).to_cif(
                separator=separator, flag_all_attributes=flag_all_attributes)
        return s_out


class SpaceGroupL(LoopN):
    """SpaceGroupL class.

    Contains information about Wyckoff positions of a space group.
    Only one site can be given for each special position but the
    remainder can be generated by applying the symmetry operations
    stored in _space_group_symop.operation_xyz.
    """

    ITEM_CLASS = SpaceGroup
    ATTR_INDEX = "id"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(SpaceGroupL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name




from typing import NoReturn
import numpy
import matplotlib.pyplot as plt
import matplotlib.widgets as wdg
import scipy
import scipy.optimize

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from .cl_1_pd_background import PdBackgroundL

class PdProc(ItemN):
    """Calculated one point for diffraction pattern.

    This section contains the diffraction data set after processing
    and application of correction terms. If the data set is
    reprocessed, this section may be replaced (with the addition of
    a new _pd_block_id entry).
    """
    ATTR_MANDATORY_NAMES = ("ttheta", )
    ATTR_MANDATORY_TYPES = (float, )
    ATTR_MANDATORY_CIF = ("2theta", )

    ATTR_OPTIONAL_NAMES = (
        "ttheta_corrected", "d_spacing", "intensity_plus_net",
        "intensity_minus_net", "intensity_plus_total", "intensity_minus_total", 
        "intensity_bkg_calc", "intensity_net", "intensity_total",
        "intensity_diff_total", "intensity_plus", "intensity_plus_sigma",
        "intensity_minus", "intensity_minus_sigma", "intensity",
        "intensity_sigma", "excluded")
    ATTR_OPTIONAL_TYPES = (float, float, float, float, float, float, float,
                           float, float, float, float, float, float, float, 
                           float, float, bool)
    ATTR_OPTIONAL_CIF = (
        "2theta_corrected", "d_spacing", "intensity_plus_net",
        "intensity_minus_net", "intensity_plus_total", "intensity_minus_total", 
        "intensity_bkg_calc", "intensity_net", "intensity_total",
        "intensity_diff_total", "intensity_plus", "intensity_plus_sigma",
        "intensity_minus", "intensity_minus_sigma", "intensity",
        "intensity_sigma", "excluded")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"ttheta": "{:.2f}",
        "ttheta_corrected": "{:.2f}", "d_spacing": "{:.5f}",
        "intensity_plus_net": "{:.2f}", "intensity_minus_net": "{:.2f}",
        "intensity_plus_total": "{:.2f}", "intensity_minus_total": "{:.2f}",
        "intensity_bkg_calc": "{:.2f}", "intensity_net": "{:.2f}",
        "intensity_total": "{:.2f}", "intensity_diff_total": "{:.2f}",
        "intensity_minus": "{:.2f}", "intensity_minus_sigma": "{:.2f}",
        "intensity": "{:.2f}", "intensity_sigma": "{:.2f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {"excluded": False}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "pd_proc"

    def __init__(self, **kwargs) -> NoReturn:
        super(PdProc, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class PdProcL(LoopN):
    """Calculated diffraction pattern.

    This section contains the diffraction data set after processing
    and application of correction terms. If the data set is
    reprocessed, this section may be replaced (with the addition of
    a new _pd_block_id entry).

    """
    ITEM_CLASS = PdProc
    ATTR_INDEX = "ttheta"
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(PdProcL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name
    
    def calc_chi_sq(self):
        pass

    def plots(self):
        return [self.plot_sum_diff()]
    
    def plot_sum_diff(self):
        """Plot unpolarized intensity vs. 2 theta
        """

        if (self.is_attribute("ttheta") and self.is_attribute("intensity_plus") and 
                self.is_attribute("intensity_plus_sigma") and
                self.is_attribute("intensity_minus") and 
                self.is_attribute("intensity_minus_sigma") and
                self.is_attribute("intensity_plus_net") and
                self.is_attribute("intensity_minus_net") and
                self.is_attribute("intensity_bkg_calc")):

            fig = plt.figure(constrained_layout=False)
            gs = fig.add_gridspec(nrows=3, ncols=1, hspace=0, height_ratios=[7,1,7])

            ax_1 = fig.add_subplot(gs[0, 0])
            ax_3 = fig.add_subplot(gs[1, 0], sharex=ax_1)
            ax_3.set_yticks([])
            ax_2 = fig.add_subplot(gs[2, 0], sharex=ax_1)

            # fig, axs = plt.subplots(nrows = 2, sharex=True)
            # ax_1, ax_2 = axs
            ax_2.set_xlabel(r"$2\theta$ (degrees)")
            ax_2.set_ylabel('Intensity (arb.u.)')
            ax_1.set_ylabel('Intensity (arb.u.)')

            np_excl = numpy.array(self.excluded, dtype=bool)
            np_notexcl = numpy.logical_not(np_excl)
            np_tth = numpy.array(self.ttheta, dtype=float)
            np_up = numpy.array(self.intensity_plus, dtype=float)
            np_sup = numpy.array(self.intensity_plus_sigma, dtype=float)
            np_up_mod = numpy.array(self.intensity_plus_net, dtype=float)
            np_down = numpy.array(self.intensity_minus, dtype=float)
            np_sdown = numpy.array(self.intensity_minus_sigma, dtype=float)
            np_down_mod = numpy.array(self.intensity_minus_net, dtype=float)
            np_bkg = numpy.array(self.intensity_bkg_calc, dtype=float)
            np_sum = np_up + np_down 
            np_sum_mod = np_up_mod + np_down_mod + np_bkg
            np_diff = np_up - np_down 
            np_diff_mod = np_up_mod - np_down_mod
            np_ssum = numpy.sqrt(numpy.square(np_sup)+numpy.square(np_sdown))
            chi_sq_points = numpy.nansum(numpy.square((np_sum - np_sum_mod)/np_ssum)[np_notexcl])/numpy.sum(np_notexcl) 
            chi_sq_diff_points = numpy.nansum(numpy.square((np_diff - np_diff_mod)/np_ssum))/np_diff_mod.size

            ax_1.set_title(f"Unpolarized ($\chi^2/n = ${chi_sq_points:.2f}) and polarized ($\chi^2/n = ${chi_sq_diff_points:.2f}) signals")
            ax_1.fill_between(np_tth, (np_sum-np_ssum), (np_sum+np_ssum), where=np_notexcl, color="k", alpha=0.4, label="experiment")
            if numpy.any(np_excl):
                ax_1.fill_between(np_tth, (np_sum-np_ssum), (np_sum+np_ssum), where=np_excl, color="r", alpha=0.5, label="excluded")
            # ax_button = plt.axes([0.25, 0.1, 0.08, 0.05])
            # grid_button = wdg.Button(ax_button, 'Grid', color='white', hovercolor='grey')
            # ax_1.errorbar(np_tth[np_notexcl], np_sum[np_notexcl], yerr=np_ssum[np_notexcl], color="k", alpha=0.2, label="experiment")
            # ax_1.errorbar(np_tth[np_excl], np_sum[np_excl], yerr=np_ssum[np_excl], color="r", alpha=0.2, label="excluded")

            y_min_d, y_max_d = ax_1.get_ylim()
            param = y_min_d-((np_sum - np_sum_mod)[np_notexcl]).max()
            coeff = np_notexcl.astype(int)
            ax_1.plot(np_tth, coeff*(np_sum - np_sum_mod)+param, "r-", alpha=0.5, label="difference")
            ax_1.plot(np_tth, np_sum_mod, "-", color="blue", label="model", linewidth=1)
            ax_1.plot([np_tth.min(), np_tth.max()], [param, param], "k:")
            ax_1.plot(np_tth, np_bkg, "b:", label="background")
            ax_1.legend(loc='upper right')

            ax_2.plot([np_tth.min(), np_tth.max()], [0., 0.], "b:")

            ax_2.fill_between(np_tth, np_diff-np_ssum, np_diff+np_ssum, color="k", alpha=0.4, label="experiment")

            # ax_2.errorbar(np_tth, np_diff, yerr=np_ssum, color="k", alpha=0.2, label="experiment") # fmt="k."

            y_min_d, y_max_d = ax_2.get_ylim()
            param = y_min_d-(np_diff-np_diff_mod).max()
            ax_2.plot([np_tth.min(), np_tth.max()], [param, param], "k:")
            ax_2.plot(np_tth, np_diff-np_diff_mod+param, "r-", alpha=0.5, label="difference")
            ax_2.plot(np_tth, np_diff_mod, "b-", label="model", linewidth=1)
            ax_2.legend(loc='upper right')

        elif (self.is_attribute("ttheta") and 
                self.is_attribute("intensity") and
                self.is_attribute("intensity_plus_net") and
                self.is_attribute("intensity_minus_net") and
                self.is_attribute("intensity_bkg_calc")):

            fig = plt.figure(constrained_layout=False)
            gs = fig.add_gridspec(nrows=2, ncols=1, hspace=0, height_ratios=[14,1])

            ax_1 = fig.add_subplot(gs[0, 0])
            ax_3 = fig.add_subplot(gs[1, 0], sharex=ax_1)
            ax_3.set_yticks([])

            ax_1.set_xlabel(r"$2\theta$ (degrees)")
            ax_1.set_ylabel('Intensity (arb.u.)')

            np_excl = numpy.array(self.excluded, dtype=bool)
            np_notexcl = numpy.logical_not(np_excl)
            np_tth = numpy.array(self.ttheta, dtype=float)
            np_sum = numpy.array(self.intensity, dtype=float)
            np_ssum = numpy.array(self.intensity_sigma, dtype=float)

            np_up_mod = numpy.array(self.intensity_plus_net, dtype=float)
            np_down_mod = numpy.array(self.intensity_minus_net, dtype=float)
            np_bkg = numpy.array(self.intensity_bkg_calc, dtype=float)
            np_sum_mod = np_up_mod + np_down_mod + np_bkg

            chi_sq_points = numpy.nansum(numpy.square((np_sum - np_sum_mod)/np_ssum)[np_notexcl])/numpy.sum(np_notexcl) 
            ax_1.set_title(f"Unpolarized signal $\chi^2/n = ${chi_sq_points:.2f}")

            ax_1.fill_between(np_tth, (np_sum-np_ssum), (np_sum+np_ssum), where=np_notexcl, color="k", alpha=0.4, label="experiment")
            ax_1.fill_between(np_tth, (np_sum-np_ssum), (np_sum+np_ssum), where=np_excl, color="r", alpha=0.5, label="excluded")

            # ax_1.errorbar(np_tth[np_notexcl], np_sum[np_notexcl], yerr=np_ssum[np_notexcl], fmt="k.", alpha=0.2, label="experiment")
            # ax_1.errorbar(np_tth[np_excl], np_sum[np_excl], yerr=np_ssum[np_excl], fmt="rs", alpha=0.2, label="excluded")

            y_min_d, y_max_d = ax_1.get_ylim()
            param = y_min_d-(np_sum - np_sum_mod).max()
            coeff = np_notexcl.astype(int)

            ax_1.plot([np_tth.min(), np_tth.max()], [param, param], "k:")
            ax_1.plot(np_tth, coeff*(np_sum - np_sum_mod)+param, "r-", alpha=0.5, label="difference")
            ax_1.plot(np_tth, np_sum_mod, "b-", label="model", linewidth=1)

            if (self.is_attribute("ttheta") and
                    self.is_attribute("intensity_bkg_calc")):
                np_tth = numpy.array(self.ttheta, dtype=float)
                np_bkg = numpy.array(self.intensity_bkg_calc, dtype=float)
                ax_1.plot(np_tth, np_bkg, "b:", label="background")
            ax_1.legend(loc='upper right')
            
        fig.tight_layout()
        return (fig, ax_1)

    def plot_diff(self):
        """Plot polarized intensity vs. 2 theta
        """

        if not(self.is_attribute("ttheta") &
               self.is_attribute("intensity_plus") & 
               self.is_attribute("intensity_plus_sigma") &
               self.is_attribute("intensity_minus") & 
               self.is_attribute("intensity_minus_sigma") &
               self.is_attribute("intensity_plus_net") &
               self.is_attribute("intensity_minus_net")):
            return 
        fig, ax = plt.subplots()
        ax.set_title("Polarized intensity: I_up - I_down")
        ax.set_xlabel("2 theta (degrees)")
        ax.set_ylabel('Intensity')
            
        np_tth = numpy.array(self.ttheta, dtype=float)
        np_up = numpy.array(self.intensity_plus, dtype=float)
        np_sup = numpy.array(self.intensity_plus_sigma, dtype=float)
        np_up_mod = numpy.array(self.intensity_plus_net, dtype=float)
        np_down = numpy.array(self.intensity_minus, dtype=float)
        np_sdown = numpy.array(self.intensity_minus_sigma, dtype=float)
        np_down_mod = numpy.array(self.intensity_minus_net, dtype=float)
        np_diff = np_up - np_down
        np_diff_mod = np_up_mod - np_down_mod
        np_sdiff = numpy.sqrt(numpy.square(np_sup)+numpy.square(np_sdown))

        chi_sq_points = numpy.nansum(numpy.square((np_diff - np_diff_mod)/np_sdiff))/(np_sdiff.size) 
        ax.set_title(f"Polarized signal $\chi^2/n = ${chi_sq_points:.2f}")

        ax.plot([np_tth.min(), np_tth.max()], [0., 0.], "b:")
        ax.fill_between(np_tth, (np_diff-np_sdiff), (np_diff+np_sdiff), color="k", alpha=0.4, label="experiment")        
        # ax.errorbar(np_tth, np_diff, yerr=np_sdiff, fmt="ko", alpha=0.2, label="experiment")
        
        y_min_d, y_max_d = ax.get_ylim()
        param = y_min_d-(np_diff-np_diff_mod).max()
        
        ax.plot([np_tth.min(), np_tth.max()], [param, param], "k:")
        ax.plot(np_tth, np_diff-np_diff_mod+param, "r-", alpha=0.5,
                    label="difference")
        ax.plot(np_tth, np_diff_mod, "b-",
                    label="model", linewidth=2)
        ax.legend(loc='upper right')
        fig.tight_layout()
        return (fig, ax)

    def estimate_background(self,
            pd_background: PdBackgroundL, offset_ttheta: float = 0):
        ttheta = numpy.array(self.ttheta, dtype=float)
        if self.is_attribute("intensity_plus"):
            signal_exp_sum = numpy.array(self.intensity_plus, dtype=float) + \
                numpy.array(self.intensity_minus, dtype=float)
            signal_exp_sum_sigma = numpy.sqrt(
                numpy.square(numpy.array(self.intensity_plus_sigma, dtype=float)) +
                numpy.square(numpy.array(self.intensity_minus_sigma, dtype=float)))
        else:
            signal_exp_sum = numpy.array(self.intensity, dtype=float)
            signal_exp_sum_sigma = numpy.array(self.intensity_sigma, dtype=float)

        signal_sum_net = numpy.array(self.intensity_plus_net, dtype=float) + \
            numpy.array(self.intensity_minus_net, dtype=float) 
        included_points = numpy.logical_not(numpy.array(self.excluded, dtype=bool))
        model_bkgr = numpy.array(self.intensity_bkg_calc, dtype=float)

        variable_names = pd_background.get_variable_names()
        param_0 = [pd_background.get_variable_by_name(name) for name in variable_names]
        def temp_func(param):
            for name, value in zip(variable_names, param):
                pd_background.set_variable_by_name(name, value)
            model_bkgr = pd_background.interpolate_by_points(ttheta+offset_ttheta)
            chi_sq_sum = numpy.nansum(numpy.square((signal_exp_sum - signal_sum_net - model_bkgr)/signal_exp_sum_sigma)[included_points])
            return chi_sq_sum
        
        res = scipy.optimize.minimize(temp_func, param_0, method="Nelder-Mead")
        
        self.numpy_intensity_bkg_calc = pd_background.interpolate_by_points(ttheta+offset_ttheta)
        self.numpy_to_items()
        return res